"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function

from .logic import Term
from collections import defaultdict


def printtrace(func):
    def _wrapped_function(*args, **kwdargs):
        print('CALL:', func.__name__, args, kwdargs)
        result = func(*args, **kwdargs)
        print('RETURN:', func.__name__, args, kwdargs, result)
        return result

    return _wrapped_function


def printtrace_self(func):
    def _wrapped_function(self, *args, **kwdargs):
        print('CALL:', func.__name__, args, kwdargs)
        result = func(self, *args, **kwdargs)
        print('RETURN:', func.__name__, args, kwdargs, result)
        return result

    return _wrapped_function


class EngineTracer(object):
    def __init__(self):
        self.call_redirect = {}
        self.call_results = defaultdict(int)
        self.level = 0

    def process_message(self, msgtype, msgtarget, msgargs, context):
        if msgtype == 'r' and msgargs[3] and msgtarget in self.call_redirect:
            self.call_return(*self.call_redirect[msgtarget])
        elif msgtype == 'c' and msgtarget in self.call_redirect:
            self.call_return(*self.call_redirect[msgtarget])

    def call_create(self, node_id, functor, context, parent):
        print('  ' * self.level, "call", Term(functor, *context))
        self.level += 1
        self.call_redirect[parent] = (node_id, functor, context)

    def call_result(self, node_id, functor, context, result):
        print('  ' * self.level, "result", Term(functor, *result))
        self.call_results[(node_id, functor, tuple(context))] += 1

    def call_return(self, node_id, functor, context):
        self.level -= 1
        if self.call_results[(node_id, functor, tuple(context))] > 0:
            # print ('  ' * self.level, "return success")
            print('  ' * self.level, "complete", Term(functor, *context))
        else:
            print('  ' * self.level, "fail", Term(functor, *context))
        self.call_results[(node_id, functor, tuple(context))] = 0

# Assume the program
#   a(1).
#   a(2).
#   p(X,Y) :- a(X), a(Y), X \= Y.
#   p(1,1).

# Call stack:
#   -> p(X,Y) == [1] a(X), [2] a(Y), [3] X \= Y
#       [1] a(X), a(Y), X \= Y
#           -> a(X)
#           <- a(1) [FACT]
#           <- a(2) [FACT]
#       [2] a(Y), 1\=X
#           -> a(Y) [CACHE]
#           <- a(1)
#           <- a(2)
#       [3] 1 \= 1
#           -> 1 \= 1
#           <- FAIL
#       [3] 1 \= 2
#           -> 1 \= 2
#           <- TRUE
#       [2] a(Y), 2 \= Y
#           -> a(Y) [CACHE]
#           <- a(1)
#           <- a(2)
#       [3] 2 \= 1
#           -> 2 \= 1
#           <- TRUE
#       [3] 2 \= 2
#           -> 2 \= 2
#           <- FAIL
#   <- p(1,2)
#   <- p(2,1)
#   <- p(1,1) [FACT]



"""
problog.maxsat - Interface to MaxSAT solvers
--------------------------------------------

Interface to MaxSAT solvers.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

from .util import mktempfile, subprocess_check_output
from . import root_path


class MaxSATSolver(object):

    def __init__(self, command):
        self.command = command

    @property
    def extension(self):
        return 'cnf'

    def prepare_input(self, formula, **kwargs):
        return formula.to_dimacs(weighted=int, **kwargs)

    def process_output(self, output):
        for line in output.split('\n'):
            if line.startswith('v '):
                return list(map(int, line.split()[1:-1]))
        return None

    def call_process(self, inputf):
        filename = mktempfile('.' + self.extension)
        with open(filename, 'w') as f:
            f.write(inputf)
        return subprocess_check_output(self.command + [filename])

    def evaluate(self, formula, **kwargs):
        inputf = self.prepare_input(formula, **kwargs)
        output = self.call_process(inputf)
        result = self.process_output(output)
        return result


class MIPMaxSATSolver(MaxSATSolver):

    def __init__(self, command):
        MaxSATSolver.__init__(self, command)

    @property
    def extension(self):
        return 'lp'

    def prepare_input(self, formula, **kwargs):
        return formula.to_lp(**kwargs)


class SCIPSolver(MIPMaxSATSolver):

    def __init__(self):
        MaxSATSolver.__init__(self, ['scip', '-f'])

    def process_output(self, output):
            facts = set()
            in_the_zone = False
            for line in output.split('\n'):
                line = line.strip()
                if line.startswith('objective value'):
                    in_the_zone = True
                elif in_the_zone:
                    if not line:
                        return list(facts)
                    else:
                        facts.add(int(line.split()[0][1:]))


def get_solver(prefer=None):
    if prefer == 'scip':
        return SCIPSolver()
    elif prefer == 'sat4j':
        return MaxSATSolver(['java', '-jar',
                             root_path('problog', 'bin', 'java', 'sat4j-maxsat.jar')])
    else:
        return MaxSATSolver(['maxsatz'])


def get_available_solvers():
    # TODO check whether they are actually available
    return ['maxsatz', 'scip', 'sat4j']



"""
problog.engine_builtin - Grounding engine builtins
--------------------------------------------------

Implementation of Prolog / ProbLog builtins.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import print_function

from .logic import term2str, Term, Clause, Constant, term2list, list2term, is_ground, is_variable
from .program import PrologFile
from .errors import GroundingError, UserError
from .engine_unify import unify_value, UnifyError, substitute_simple
from .engine import UnknownClauseInternal, UnknownClause

import os
import imp  # For load_external
import inspect  # For load_external


def add_standard_builtins(engine, b=None, s=None, sp=None):
    """Adds standard builtins to the given engine.

    :param engine: engine to add builtins to
    :type engine: ClauseDBEngine
    :param b: wrapper for boolean builtins (returning True/False)
    :param s: wrapper for simple builtins (return deterministic results)
    :param sp: wrapper for probabilistic builtins (return probabilistic results)
    """

    engine.add_builtin('true', 0, b(_builtin_true))  # -1
    engine.add_builtin('fail', 0, b(_builtin_fail))  # -2
    engine.add_builtin('false', 0, b(_builtin_fail))  # -3

    engine.add_builtin('=', 2, s(_builtin_eq))  # -4
    engine.add_builtin('\=', 2, b(_builtin_neq))  # -5

    engine.add_builtin('findall', 3, sp(_builtin_findall))  # -6
    engine.add_builtin('findall_top', 3, sp(_builtin_findall_top))  # -7

    engine.add_builtin('==', 2, b(_builtin_same))
    engine.add_builtin('\==', 2, b(_builtin_notsame))

    engine.add_builtin('is', 2, s(_builtin_is))

    engine.add_builtin('>', 2, b(_builtin_gt))
    engine.add_builtin('<', 2, b(_builtin_lt))
    engine.add_builtin('=<', 2, b(_builtin_le))
    engine.add_builtin('>=', 2, b(_builtin_ge))
    engine.add_builtin('=\=', 2, b(_builtin_val_neq))
    engine.add_builtin('=:=', 2, b(_builtin_val_eq))

    engine.add_builtin('var', 1, b(_builtin_var))
    engine.add_builtin('atom', 1, b(_builtin_atom))
    engine.add_builtin('atomic', 1, b(_builtin_atomic))
    engine.add_builtin('compound', 1, b(_builtin_compound))
    engine.add_builtin('float', 1, b(_builtin_float))
    engine.add_builtin('rational', 1, b(_builtin_rational))
    engine.add_builtin('integer', 1, b(_builtin_integer))
    engine.add_builtin('nonvar', 1, b(_builtin_nonvar))
    engine.add_builtin('number', 1, b(_builtin_number))
    engine.add_builtin('simple', 1, b(_builtin_simple))
    engine.add_builtin('callable', 1, b(_builtin_callable))
    engine.add_builtin('dbreference', 1, b(_builtin_dbreference))
    engine.add_builtin('primitive', 1, b(_builtin_primitive))
    engine.add_builtin('ground', 1, b(_builtin_ground))
    engine.add_builtin('is_list', 1, b(_builtin_is_list))

    engine.add_builtin('=..', 2, s(_builtin_split_call))
    engine.add_builtin('arg', 3, s(_builtin_arg))
    engine.add_builtin('functor', 3, s(_builtin_functor))

    engine.add_builtin('@>', 2, b(_builtin_struct_gt))
    engine.add_builtin('@<', 2, b(_builtin_struct_lt))
    engine.add_builtin('@>=', 2, b(_builtin_struct_ge))
    engine.add_builtin('@=<', 2, b(_builtin_struct_le))
    engine.add_builtin('compare', 3, s(_builtin_compare))

    engine.add_builtin('length', 2, s(_builtin_length))
    # engine.add_builtin('call_external', 2, s(_builtin_call_external))

    engine.add_builtin('sort', 2, s(_builtin_sort))
    engine.add_builtin('between', 3, s(_builtin_between))
    engine.add_builtin('succ', 2, s(_builtin_succ))
    engine.add_builtin('plus', 3, s(_builtin_plus))

    engine.add_builtin('consult', 1, b(_builtin_consult))
    engine.add_builtin('.', 2, b(_builtin_consult_as_list))
    # engine.add_builtin('load_external', 1, b(_builtin_load_external))
    engine.add_builtin('unknown', 1, b(_builtin_unknown))

    engine.add_builtin('use_module', 1, b(_builtin_use_module))

    engine.add_builtin('call', 1, _builtin_call)
    for i in range(2, 10):
        engine.add_builtin('call', i, _builtin_calln)

    engine.add_builtin('subquery', 2, s(_builtin_subquery))
    engine.add_builtin('subquery', 3, s(_builtin_subquery))

    engine.add_builtin('sample_uniform1', 3, sp(_builtin_sample_uniform))

    for i in range(1, 10):
        engine.add_builtin('debugprint', i, b(_builtin_debugprint))

    for i in range(1, 10):
        engine.add_builtin('write', i, b(_builtin_write))

    for i in range(1, 10):
        engine.add_builtin('writenl', i, b(_builtin_writenl))

    for i in range(1, 10):
        engine.add_builtin('error', i, b(_builtin_error))

    engine.add_builtin('nl', 0, b(_builtin_nl))



# noinspection PyUnusedLocal
def _builtin_debugprint(*args, **kwd):
    print(' '.join(map(term2str, args)))
    return True


def term2str_noquote(term):
    res = term2str(term)
    if res[0] == res[-1] == "'":
        res = res[1:-1]
    return res

def _builtin_write(*args, **kwd):
    print(' '.join(map(term2str_noquote, args)), end='')
    return True


def _builtin_error(*args, **kwd):
    location = kwd.get('call_origin', (None, None))[1]
    database = kwd['database']
    location = database.lineno(location)
    message = ''.join(map(term2str_noquote, args))
    raise UserError(message, location=location)


def _builtin_writenl(*args, **kwd):
    print(' '.join(map(term2str_noquote, args)))
    return True

def _builtin_nl(**kwd):
    print()
    return True



class CallModeError(GroundingError):
    """
    Represents an error in builtin argument types.
    """

    def __init__(self, functor, args, accepted=None, message=None, location=None):
        if accepted is None:
            accepted = []
        if functor:
            self.scope = '%s/%s' % (functor, len(args))
        else:
            self.scope = None
        self.received = ', '.join(map(self._show_arg, args))
        self.expected = [', '.join(map(self._show_mode, mode)) for mode in accepted]
        msg = 'Invalid argument types for call'
        if self.scope:
            msg += " to '%s'" % self.scope
        msg += ': arguments: (%s)' % self.received
        if accepted:
            msg += ', expected: (%s)' % ') or ('.join(self.expected)
        else:
            msg += ', expected: ' + message
        GroundingError.__init__(self, msg, location)

    def _show_arg(self, x):
        return term2str(x)

    def _show_mode(self, t):
        return mode_types[t][0]


class StructSort(object):
    """
    Comparator of terms based on structure.
    """

    # noinspection PyUnusedLocal
    def __init__(self, obj, *args):
        self.obj = obj

    def __lt__(self, other):
        return struct_cmp(self.obj, other.obj) < 0

    def __gt__(self, other):
        return struct_cmp(self.obj, other.obj) > 0

    def __eq__(self, other):
        return struct_cmp(self.obj, other.obj) == 0

    def __le__(self, other):
        return struct_cmp(self.obj, other.obj) <= 0

    def __ge__(self, other):
        return struct_cmp(self.obj, other.obj) >= 0

    def __ne__(self, other):
        return struct_cmp(self.obj, other.obj) != 0


def _is_var(term):
    return is_variable(term) or term.is_var()


def _is_nonvar(term):
    return not _is_var(term)


def _is_term(term):
    return not _is_var(term) and not _is_constant(term)


def _is_float_pos(term):
    return _is_constant(term) and term.is_float()


def _is_float_neg(term):
    return _is_term(term) and term.arity == 1 and term.functor == "'-'" and \
        _is_float_pos(term.args[0])


def _is_float(term):
    return _is_float_pos(term) or _is_float_neg(term)


def _is_integer_pos(term):
    return _is_constant(term) and term.is_integer()


def _is_integer_neg(term):
    return _is_term(term) and term.arity == 1 and term.functor == "'-'" and \
        _is_integer_pos(term.args[0])


def _is_integer(term):
    return _is_integer_pos(term) or _is_integer_neg(term)


def _is_string(term):
    return _is_constant(term) and term.is_string()


def _is_number(term):
    return _is_float(term) or _is_integer(term)


def _is_constant(term):
    return not _is_var(term) and term.is_constant()


def _is_atom(term):
    return _is_term(term) and term.arity == 0


def _is_atomic(term):
    return _is_nonvar(term) and not _is_compound(term)


# noinspection PyUnusedLocal
def _is_rational(term):
    return False


# noinspection PyUnusedLocal
def _is_dbref(term):
    return False


def _is_compound(term):
    return _is_term(term) and term.arity > 0


def _is_list_maybe(term):
    """
    Check whether the term looks like a list (i.e. of the form '.'(_,_)).
    :param term:
    :return:
    """
    return _is_compound(term) and term.functor == '.' and term.arity == 2


def _is_list_nonempty(term):
    if _is_list_maybe(term):
        tail = list_tail(term)
        return _is_list_empty(tail) or _is_var(tail)
    return False


def _is_fixed_list(term):
    return _is_list_empty(term) or _is_fixed_list_nonempty(term)


def _is_fixed_list_nonempty(term):
    if _is_list_maybe(term):
        tail = list_tail(term)
        return _is_list_empty(tail)
    return False


def _is_list_empty(term):
    return _is_atom(term) and term.functor == '[]'


def _is_list(term):
    return _is_list_empty(term) or _is_list_nonempty(term)


def _is_compare(term):
    return _is_atom(term) and term.functor in ("'<'", "'='", "'>'")


mode_types = {
    'i': ('integer', _is_integer),
    'I': ('positive_integer', _is_integer_pos),
    'f': ('float', _is_float),
    'v': ('var', _is_var),
    'n': ('nonvar', _is_nonvar),
    'l': ('list', _is_list),
    'L': ('fixed_list', _is_fixed_list),  # List of fixed length (i.e. tail is [])
    '*': ('any', lambda x: True),
    '<': ('compare', _is_compare),  # < = >
    'g': ('ground', is_ground),
    'a': ('atom', _is_atom),
    'c': ('callable', _is_term)
}


# noinspection PyUnusedLocal
def check_mode(args, accepted, functor=None, location=None, database=None, **kwdargs):
    """Checks the arguments against a list of accepted types.

    :param args: arguments to check
    :type args: tuple of Term
    :param accepted: list of accepted combination of types (see mode_types)
    :type accepted: list of str
    :param functor: functor of the call (used for error message)
    :param location: location of the call (used for error message)
    :param database: database (used for error message)
    :param kwdargs: additional arguments (not used)
    :return: the index of the first mode in accepted that matches the arguments
    :rtype: int
    """
    for i, mode in enumerate(accepted):
        correct = True
        for a, t in zip(args, mode):
            name, test = mode_types[t]
            if not test(a):
                correct = False
                break
        if correct:
            return i
    if database and location:
        location = database.lineno(location)
    else:
        location = None
    raise CallModeError(functor, args, accepted, location=location)


def list_elements(term):
    """Extract elements from a List term.
    Ignores the list tail.

    :param term: term representing a list
    :type term: Term
    :return: elements of the list
    :rtype: list of Term
    """
    elements = []
    tail = term
    while _is_list_maybe(tail):
        elements.append(tail.args[0])
        tail = tail.args[1]
    return elements, tail


def list_tail(term):
    """Extract the tail of the list.

    :param term: Term representing a list
    :type term: Term
    :return: tail of the list
    :rtype: Term
    """
    tail = term
    while _is_list_maybe(tail):
        tail = tail.args[1]
    return tail


def _builtin_split_call(term, parts, database=None, location=None, **kwdargs):
    """Implements the '=..' builtin operator.

    :param term:
    :param parts:
    :param database:
    :param location:
    :param kwdargs:
    :return:
    """
    functor = '=..'
    # modes:
    #   <v> =.. list  => list has to be fixed length and non-empty
    #                       IF its length > 1 then first element should be an atom
    #   <n> =.. <list or var>
    #
    mode = check_mode((term, parts), ['vL', 'nv', 'nl'], functor=functor, **kwdargs)
    if mode == 0:
        elements, tail = list_elements(parts)
        if len(elements) == 0:
            raise CallModeError(functor, (term, parts),
                                message='non-empty list for arg #2 if arg #1 is a variable',
                                location=database.lineno(location))
        elif len(elements) > 1 and not _is_atom(elements[0]):
            raise CallModeError(functor, (term, parts),
                                message='atom as first element in list if arg #1 is a variable',
                                location=database.lineno(location))
        elif len(elements) == 1:
            # Special case => term == parts[0]
            return [(elements[0], parts)]
        else:
            term_part = elements[0](*elements[1:])
            return [(term_part, parts)]
    else:
        part_list = (term.with_args(),) + term.args
        current = Term('[]')
        for t in reversed(part_list):
            current = Term('.', t, current)
        try:
            local_values = {}
            list_part = unify_value(current, parts, local_values)
            elements, tail = list_elements(list_part)
            term_new = elements[0](*elements[1:])
            term_part = unify_value(term, term_new, local_values)
            return [(term_part, list_part)]
        except UnifyError:
            return []


def _builtin_arg(index, term, arguments, **kwdargs):
    check_mode((index, term, arguments), ['In*'], functor='arg', **kwdargs)
    index_v = int(index) - 1
    if 0 <= index_v < len(term.args):
        try:
            arg = term.args[index_v]
            res = unify_value(arg, arguments, {})
            return [(index, term, res)]
        except UnifyError:
            pass
    return []


def _builtin_functor(term, functor, arity, **kwdargs):
    mode = check_mode((term, functor, arity), ['vaI', 'n**'], functor='functor', **kwdargs)

    if mode == 0:
        kwdargs.get('callback').newResult(Term(functor, *((None,) * int(arity))), functor, arity)
    else:
        try:
            values = {}
            func_out = unify_value(functor, Term(term.functor), values)
            arity_out = unify_value(arity, Constant(term.arity), values)
            return [(term, func_out, arity_out)]
        except UnifyError:
            pass
    return []


# noinspection PyUnusedLocal
def _builtin_true(**kwdargs):
    """``true``"""
    return True


# noinspection PyUnusedLocal
def _builtin_fail(**kwdargs):
    """``fail``"""
    return False


# noinspection PyUnusedLocal
def _builtin_eq(arg1, arg2, **kwdargs):
    """``A = B``
        A and B not both variables
    """
    try:
        result = unify_value(arg1, arg2, {})
        return [(result, result)]
    except UnifyError:
        return []
        # except VariableUnification:
        #     raise VariableUnification(location = database.lineno(location))


# noinspection PyUnusedLocal
def _builtin_neq(arg1, arg2, **kwdargs):
    """``A \= B``
        A and B not both variables
    """
    try:
        unify_value(arg1, arg2, {})
        return False
    except UnifyError:
        return True


# noinspection PyUnusedLocal
def _builtin_notsame(arg1, arg2, **kwdargs):
    """``A \== B``"""
    return not arg1 == arg2


# noinspection PyUnusedLocal
def _builtin_same(arg1, arg2, **kwdargs):
    """``A == B``"""
    return arg1 == arg2


def _builtin_gt(arg1, arg2, engine=None, **kwdargs):
    """``A > B``
        A and B are ground
    """
    check_mode((arg1, arg2), ['gg'], functor='>', **kwdargs)
    a_value = arg1.compute_value(engine.functions)
    b_value = arg2.compute_value(engine.functions)
    if a_value is None or b_value is None:
        return False
    else:
        return a_value > b_value


def _builtin_lt(arg1, arg2, engine=None, **kwdargs):
    """``A > B``
        A and B are ground
    """
    check_mode((arg1, arg2), ['gg'], functor='<', **kwdargs)
    a_value = arg1.compute_value(engine.functions)
    b_value = arg2.compute_value(engine.functions)
    if a_value is None or b_value is None:
        return False
    else:
        return a_value < b_value


def _builtin_le(arg1, arg2, engine=None, **k):
    """``A =< B``
        A and B are ground
    """
    check_mode((arg1, arg2), ['gg'], functor='=<', **k)
    a_value = arg1.compute_value(engine.functions)
    b_value = arg2.compute_value(engine.functions)
    if a_value is None or b_value is None:
        return False
    else:
        return a_value <= b_value


def _builtin_ge(arg1, arg2, engine=None, **k):
    """``A >= B``
        A and B are ground
    """
    check_mode((arg1, arg2), ['gg'], functor='>=', **k)
    a_value = arg1.compute_value(engine.functions)
    b_value = arg2.compute_value(engine.functions)
    if a_value is None or b_value is None:
        return False
    else:
        return a_value >= b_value


def _builtin_val_neq(a, b, engine=None, **k):
    """``A =\= B``
        A and B are ground
    """
    check_mode((a, b), ['gg'], functor='=\=', **k)
    a_value = a.compute_value(engine.functions)
    b_value = b.compute_value(engine.functions)
    if a_value is None or b_value is None:
        return False
    else:
        return a_value != b_value


def _builtin_val_eq(a, b, engine=None, **k):
    """``A =:= B``
        A and B are ground
    """
    check_mode((a, b), ['gg'], functor='=:=', **k)
    a_value = a.compute_value(engine.functions)
    b_value = b.compute_value(engine.functions)
    if a_value is None or b_value is None:
        return False
    else:
        return a_value == b_value


def _builtin_is(a, b, engine=None, **k):
    """``A is B``
        B is ground

        @param a:
        @param b:
        @param engine:
        @param k:
    """
    check_mode((a, b), ['*g'], functor='is', **k)
    try:
        b_value = b.compute_value(engine.functions)
        if b_value is None:
            return []
        else:
            r = Constant(b_value)
            unify_value(a, r, {})
            return [(r, b)]
    except UnifyError:
        return []


# noinspection PyUnusedLocal
def _builtin_var(term, **k):
    return _is_var(term)


# noinspection PyUnusedLocal
def _builtin_atom(term, **k):
    return _is_atom(term)


# noinspection PyUnusedLocal
def _builtin_atomic(term, **k):
    return _is_atom(term) or _is_number(term)


# noinspection PyUnusedLocal
def _builtin_compound(term, **k):
    return _is_compound(term)


# noinspection PyUnusedLocal
def _builtin_float(term, **k):
    return _is_float(term)


# noinspection PyUnusedLocal
def _builtin_integer(term, **k):
    return _is_integer(term)


# noinspection PyUnusedLocal
def _builtin_nonvar(term, **k):
    return not _is_var(term)


# noinspection PyUnusedLocal
def _builtin_number(term, **k):
    return _is_number(term)


# noinspection PyUnusedLocal
def _builtin_simple(term, **k):
    return _is_var(term) or _is_atomic(term)


# noinspection PyUnusedLocal
def _builtin_callable(term, **k):
    return _is_term(term)


# noinspection PyUnusedLocal
def _builtin_rational(term, **k):
    return _is_rational(term)


# noinspection PyUnusedLocal
def _builtin_dbreference(term, **k):
    return _is_dbref(term)


# noinspection PyUnusedLocal
def _builtin_primitive(term, **k):
    return _is_atomic(term) or _is_dbref(term)


# noinspection PyUnusedLocal
def _builtin_ground(term, **k):
    return is_ground(term)


# noinspection PyUnusedLocal
def _builtin_is_list(term, **k):
    return _is_list(term)


def compare(a, b):
    if a < b:
        return -1
    elif a > b:
        return 1
    else:
        return 0


def struct_cmp(a, b):
    # Note: structural comparison
    # 1) Var < Num < Str < Atom < Compound
    # 2) Var by address
    # 3) Number by value, if == between int and float => float is smaller
    #   (iso prolog: Float always < Integer )
    # 4) String alphabetical
    # 5) Atoms alphabetical
    # 6) Compound: arity / functor / arguments

    # 1) Variables are smallest
    if _is_var(a):
        if _is_var(b):
            # 2) Variable by address
            return compare(a, b)
        else:
            return -1
    elif _is_var(b):
        return 1
    # assert( not is_var(A) and not is_var(B) )

    # 2) Numbers are second smallest
    if _is_number(a):
        if _is_number(b):
            # Just compare numbers on float value
            res = compare(float(a), float(b))
            if res == 0:
                # If the same, float is smaller.
                if _is_float(a) and _is_integer(b):
                    return -1
                elif _is_float(b) and _is_integer(a):
                    return 1
                else:
                    return 0
        else:
            return -1
    elif _is_number(b):
        return 1

    # 3) Strings are third
    if _is_string(a):
        if _is_string(b):
            return compare(str(a), str(b))
        else:
            return -1
    elif _is_string(b):
        return 1

    # 4) Atoms / terms come next
    # 4.1) By arity
    res = compare(a.arity, b.arity)
    if res != 0:
        return res

    # 4.2) By functor
    res = compare(a.functor, b.functor)
    if res != 0:
        return res

    # 4.3) By arguments (recursively)
    for a1, b1 in zip(a.args, b.args):
        res = struct_cmp(a1, b1)
        if res != 0:
            return res

    return 0


# noinspection PyUnusedLocal
def _builtin_struct_lt(a, b, **k):
    return struct_cmp(a, b) < 0


# noinspection PyUnusedLocal
def _builtin_struct_le(a, b, **k):
    return struct_cmp(a, b) <= 0


# noinspection PyUnusedLocal
def _builtin_struct_gt(a, b, **k):
    return struct_cmp(a, b) > 0


# noinspection PyUnusedLocal
def _builtin_struct_ge(a, b, **k):
    return struct_cmp(a, b) >= 0


def _builtin_compare(c, a, b, **k):
    mode = check_mode((c, a, b), ['<**', 'v**'], functor='compare', **k)
    compares = "'>'", "'='", "'<'"
    cp = struct_cmp(a, b)
    c_token = compares[1 - cp]

    if mode == 0:  # Given compare
        if c_token == c.functor:
            return [(c, a, b)]
    else:  # Unknown compare
        return [(Term(c_token), a, b)]


# numbervars(T,+N1,-Nn)    number the variables TBD?

def build_list(elements, tail):
    current = tail
    for el in reversed(elements):
        current = Term('.', el, current)
    return current


# class UnknownExternal(GroundingError):
#     """Undefined clause in call."""
#
#     def __init__(self, signature, location):
#         GroundingError.__init__(self, "Unknown external function '%s'" % signature, location)


# def _builtin_call_external(call, result, database=None, location=None, **k):
#     from . import pypl
#     check_mode((call, result), ['gv'], function='call_external', database=database,
#                location=location, **k)
#
#     func = k['engine'].get_external_call(call.functor)
#     if func is None:
#         raise UnknownExternal(call.functor, database.lineno(location))
#
#     values = [pypl.pl2py(arg) for arg in call.args]
#     computed_result = func(*values)
#
#     return [(call, pypl.py2pl(computed_result))]


def _builtin_length(l, n, **k):
    mode = check_mode((l, n), ['LI', 'Lv', 'lI', 'vI'], functor='length', **k)
    # Note that Prolog also accepts 'vv' and 'lv', but these are unbounded.
    # Note that lI is a subset of LI, but only first matching mode is returned.
    if mode == 0 or mode == 1:  # Given fixed list and maybe length
        elements, tail = list_elements(l)
        list_size = len(elements)
        try:
            n = unify_value(n, Constant(list_size), {})
            return [(l, n)]
        except UnifyError:
            return []
    else:  # Unbounded list or variable list and fixed length.
        if mode == 2:
            elements, tail = list_elements(l)
        else:
            elements, tail = [], l
        remain = int(n) - len(elements)
        if remain < 0:
            raise UnifyError()
        else:
            min_var = k.get('engine')._context_min_var(k.get('context'))
            extra = list(range(min_var, min_var - remain, -1))  # [None] * remain
        new_l = build_list(elements + extra, Term('[]'))
        return [(new_l, n)]


def _builtin_sort(l, s, **k):
    # TODO doesn't work properly with variables e.g. gives sort([X,Y,Y],[_])
    # should be sort([X,Y,Y],[X,Y])
    check_mode((l, s), ['L*'], functor='sort', **k)
    elements, tail = list_elements(l)
    # assert( is_list_empty(tail) )
    try:
        sorted_list = build_list(sorted(set(elements), key=StructSort), Term('[]'))
        s_out = unify_value(s, sorted_list, {})
        return [(l, s_out)]
    except UnifyError:
        return []


def _builtin_between(low, high, value, **k):
    """
    Implements the between/3 builtin.
   :param low:
   :param high:
   :param value:
   :param k:
   :return:
    """
    mode = check_mode((low, high, value), ['iii', 'iiv'], functor='between', **k)
    low_v = int(low)
    high_v = int(high)
    if mode == 0:  # Check
        value_v = int(value)
        if low_v <= value_v <= high_v:
            return [(low, high, value)]
    else:  # Enumerate
        results = []
        for value_v in range(low_v, high_v + 1):
            results.append((low, high, Constant(value_v)))
        return results


def _builtin_succ(a, b, **kwdargs):
    """
    Implements the succ/2 builtin.
   :param a: input argument
   :param b: output argument
   :param kwdargs: additional arguments
   :return:
    """
    mode = check_mode((a, b), ['vI', 'Iv', 'II'], functor='succ', **kwdargs)
    if mode == 0:
        b_v = int(b)
        return [(Constant(b_v - 1), b)]
    elif mode == 1:
        a_v = int(a)
        return [(a, Constant(a_v + 1))]
    else:
        a_v = int(a)
        b_v = int(b)
        if b_v == a_v + 1:
            return [(a, b)]
    return []


def _builtin_plus(a, b, c, **kwdargs):
    """
    Implements the plus/3 builtin.
   :param a: first argument
   :param b: second argument
   :param c: result argument
   :param kwdargs: additional arguments
   :return:
    """
    mode = check_mode((a, b, c), ['iii', 'iiv', 'ivi', 'vii'], functor='plus', **kwdargs)
    if mode == 0:
        a_v = int(a)
        b_v = int(b)
        c_v = int(c)
        if a_v + b_v == c_v:
            return [(a, b, c)]
    elif mode == 1:
        a_v = int(a)
        b_v = int(b)
        return [(a, b, Constant(a_v + b_v))]
    elif mode == 2:
        a_v = int(a)
        c_v = int(c)
        return [(a, Constant(c_v - a_v), c)]
    else:
        b_v = int(b)
        c_v = int(c)
        return [(Constant(c_v - b_v), b, c)]
    return []


def _atom_to_filename(atom):
    """Translate an atom to a filename.

   :param atom: filename as atom
   :type atom: Term
   :return: filename as string
   :rtype: str
    """
    atomstr = str(atom)
    if atomstr[0] == atomstr[-1] == "'":
        atomstr = atomstr[1:-1]
    return atomstr


class ConsultError(GroundingError):
    """Error during consult"""

    def __init__(self, message, location):
        GroundingError.__init__(self, message, location)


def _builtin_consult_as_list(op1, op2, **kwdargs):
    """Implementation of consult/1 using list notation.

   :param op1: first element in the list
   :param op2: tail of the list
   :param kwdargs: additional arugments
   :return: True
    """
    # TODO make non-recursive
    check_mode((op1, op2), ['*L'], functor='consult', **kwdargs)
    _builtin_consult(op1, **kwdargs)
    if _is_list_nonempty(op2):
        _builtin_consult_as_list(op2.args[0], op2.args[1], **kwdargs)
    return True


def _builtin_consult(filename, database=None, engine=None, **kwdargs):
    """
    Implementation of consult/1 builtin.
    A file will be loaded only once.
   :param filename: filename to load into the database
   :type filename: Term
   :param database: database containing the current logic program.
   :param kwdargs: additional arguments
   :return: True
    """

    root = database.source_root
    if filename.location:
        source_root = database.source_files[filename.location[0]]
        if source_root:
            root = os.path.dirname(source_root)
    check_mode((filename,), ['a'], functor='consult', **kwdargs)
    filename = os.path.join(root, _atom_to_filename(filename))
    if not os.path.exists(filename):
        filename += '.pl'
    if not os.path.exists(filename):
        raise ConsultError(message="Consult: file not found '%s'" % filename,
                           location=database.lineno(kwdargs.get('location')))

    # Prevent loading the same file twice
    if filename not in database.source_files:
        identifier = len(database.source_files)
        database.source_files.append(filename)
        database.source_parent.append(kwdargs.get('location'))
        pl = PrologFile(filename, identifier=identifier)
        database.line_info.append(pl.line_info[0])
        for clause in pl:
            database += clause
        # engine._process_directives(database)

    return True


# # noinspection PyUnusedLocal
# def _builtin_load_external(arg, engine=None, database=None, location=None, **kwdargs):
#     check_mode((arg,), ['a'], functor='load_external')
#     # Load external (python) files that are referenced in the model
#     externals = {}
#     root = database.source_root
#     if arg.location:
#         root = os.path.dirname(database.source_files[arg.location[0]])
#     filename = os.path.join(root, _atom_to_filename(arg))
#     if not os.path.exists(filename):
#         raise ConsultError(message="Load external: file not found '%s'" % filename,
#                            location=database.lineno(location))
#     try:
#         with open(filename, 'r') as extfile:
#             ext = imp.load_module('externals', extfile, filename, ('.py', 'U', 1))
#             for func_name, func in inspect.getmembers(ext, inspect.isfunction):
#                 externals[func_name] = func
#         engine.add_external_calls(externals)
#     except ImportError:
#         raise ConsultError(message="Error while loading external file '%s'" % filename,
#                            location=database.lineno(location))
#
#     return True


# noinspection PyUnusedLocal
def _builtin_unknown(arg, engine=None, **kwdargs):
    check_mode((arg,), ['a'], functor='unknown')
    if arg.functor == 'fail':
        engine.unknown = engine.UNKNOWN_FAIL
    else:
        engine.unknown = engine.UNKNOWN_ERROR
    return True


def _select_sublist(lst, target):
    """
    Enumerate all possible selection of elements from a list.
    This function is used to generate all solutions to findall/3.
    An element must be selected if it is TRUE in the target formula.

    :param lst: list to select elements from
    :type lst: list of tuple
    :param target: data structure containing truth value of nodes
    :type target: LogicFormula
    :return: generator of sublists
    """
    l = len(lst)

    # Generate an array that indicates the decision bit for each element in the list.
    # If an element is deterministically true, then no decision bit is needed.
    choice_bits = [None] * l
    x = 0
    for i in range(0, l):
        if lst[i][1] != target.TRUE:
            choice_bits[i] = x
            x += 1
    # We have 2^x distinct lists. Each can be represented as a number between 0 and 2^x-1=n.
    n = (1 << x) - 1

    while n >= 0:
        # Generate the list of positive values and node identifiers
        # noinspection PyTypeChecker
        sublist = [lst[i] for i in range(0, l) if choice_bits[i] is None or n & 1 << choice_bits[i]]
        # Generate the list of negative node identifiers
        # noinspection PyTypeChecker
        sublist_no = tuple([-lst[i][1] for i in range(0, l) if
                            not (choice_bits[i] is None or n & 1 << choice_bits[i])])
        if sublist:
            terms, nodes = zip(*sublist)
        else:
            # Empty list.
            terms, nodes = (), ()
        yield terms, nodes + sublist_no + (0,)
        n -= 1


def _builtin_findall_base(pattern, goal, result, top_only=False, database=None, target=None,
                          engine=None, context=None, **kwdargs):
    """
    Implementation of findall/3 builtin.
   :param pattern: pattern to extract
   :type pattern: Term
   :param goal: goal to evaluate
   :type goal: Term
   :param result: list to store results
   :type result: Term
   :param database: database holding logic program
   :type database: ClauseDB
   :param target: logic formula in which to store the result
   :type target: LogicFormula
   :param engine: engine that is used for evaluation
   :type engine: ClauseDBEngine
   :param kwdargs: additional arguments from engine
   :return: list results (tuple of lists and node identifiers)
    """
    # Check the modes.
    mode = check_mode((pattern, goal, result,), ['*cv', '*cl'], database=database, **kwdargs)

    findall_head = Term(engine.get_non_cache_functor(), pattern, *goal.variables())
    findall_clause = Clause(findall_head, goal)
    findall_db = database.extend()
    findall_db += findall_clause

    class _TranslateToNone(object):

        # noinspection PyUnusedLocal
        def __getitem__(self, item):
            return None

    findall_head = substitute_simple(findall_head, _TranslateToNone())

    results = engine.call(findall_head, subcall=True, database=findall_db, target=target, **kwdargs)
    results = [(res[0], n) for res, n in results]
    output = []
    if top_only:
        if results:
            # Only return the maximal list.
            l, n = zip(*results)
            node = target.add_and(n)
            if node is not None:
                res = build_list(l, Term('[]'))
                if mode == 0:  # var
                    args = (pattern, goal, res)
                    output.append((args, node))
                else:
                    try:
                        res = unify_value(res, result, {})
                        args = (pattern, goal, res)
                        output.append((args, node))
                    except UnifyError:
                        pass
    else:
        for l, n in _select_sublist(results, target):
            node = target.add_and(n)
            if node is not None:
                res = build_list(l, Term('[]'))
                if mode == 0:  # var
                    args = (pattern, goal, res)
                    output.append((args, node))
                else:
                    try:
                        res = unify_value(res, result, {})
                        args = (pattern, goal, res)
                        output.append((args, node))
                    except UnifyError:
                        pass
    return output


# noinspection PyUnusedLocal
def _builtin_sample_all(pattern, goal, result, database=None, target=None, **kwdargs):
    # Like findall.
    pass


def _builtin_sample_uniform(key, lst, result, database=None, target=None, **kwdargs):
    """Implements the sample_uniform(+Key,+List,-Result) builtin.
    This predicate succeeds once for each element in the list as result, and with probability \
    1/(length of the list).
    The first argument is used as an identifier such that calls with the same key enforce mutual \
    exclusivity on the results, that is, the probability of

        sample_uniform(K,L,R1), sample_uniform(K,L,R2), R1 \== R2

    is 0.



    :param key:
    :param lst:
    :param result:
    :param database:
    :type database: StackBasedEngine
    :param target:
    :type target: LogicFormula
    :param kwdargs:
    :return:
    """
    mode = check_mode((key, lst, result,), ['gLv', 'gLn'], database=database, **kwdargs)
    identifier = '_uniform_%s' % key
    elements, tail = list_elements(lst)
    if len(elements) == 0:
        return []
    else:
        prob = Constant(1 / float(len(elements)))
        results = []
        if mode == 0:
            for i, elem in enumerate(elements):
                elem_identifier = (identifier, i)
                # res = unify_value(result, elem)
                results.append(((key, lst, elem),
                                target.add_atom(identifier=elem_identifier, probability=prob,
                                                group=identifier)))
        else:
            res = None
            for el in elements:
                try:
                    res = unify_value(el, result, {})
                    break
                except UnifyError:
                    pass
            if res is not None:
                results.append(((key, lst, res),
                                target.add_atom(identifier=identifier, probability=prob)))
        return results


def _builtin_findall(pattern, goal, result, **kwdargs):
    return _builtin_findall_base(pattern, goal, result, top_only=False, **kwdargs)


def _builtin_findall_top(pattern, goal, result, **kwdargs):
    return _builtin_findall_base(pattern, goal, result, top_only=True, **kwdargs)


# noinspection PyPep8Naming
class problog_export(object):
    database = None

    @classmethod
    def add_function(cls, name, in_args, out_args, function):
        if problog_export.database is not None:
            problog_export.database.add_extern(name, in_args + out_args, function)

    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwdargs):
        # TODO check if arguments are in order: input first, output last
        self.input_arguments = [a[1:] for a in args if a[0] == '+']
        self.output_arguments = [a[1:] for a in args if a[0] == '-']

    def _convert_input(self, a, t):
        if t == 'str':
            return str(a)
        elif t == 'int':
            return int(a)
        elif t == 'float':
            return float(a)
        elif t == 'list':
            return term2list(a)
        elif t == 'term':
            return a
        else:
            raise ValueError("Unknown type specifier '%s'!" % t)

    def _type_to_callmode(self, t):
        if t == 'str':
            return 'a'
        elif t == 'int':
            return 'i'
        elif t == 'float':
            return 'f'
        elif t == 'list':
            return 'L'
        elif t == 'term':
            return '*'
        else:
            raise ValueError("Unknown type specifier '%s'!" % t)

    def _extract_callmode(self):
        callmode_in = ''
        for t in self.input_arguments:
            callmode_in += self._type_to_callmode(t)

        # multiple call modes: index = binary encoding on whether the output is bound
        # 0 -> all unbound
        # 1 -> first output arg is bound
        # 2 -> second output arg is bound
        # 3 -> first and second are bound

        n = len(self.output_arguments)
        for i in range(0, 1 << n):
            callmode = callmode_in
            for j, t in enumerate(self.output_arguments):
                if i & (1 << (n - j - 1)):
                    callmode += self._type_to_callmode(t)
                else:
                    callmode += 'v'
            yield callmode

    def _convert_output(self, a, t):
        if t == 'str':
            return Term(a)
        elif t == 'int':
            return Constant(a)
        elif t == 'float':
            return Constant(a)
        elif t == 'list':
            return list2term(a)
        elif t == 'term':
            if not isinstance(a, Term):
                raise ValueError("Expected term output, got '%s' instead." % type(a))
            return a
        else:
            raise ValueError("Unknown type specifier '%s'!" % t)

    def _convert_inputs(self, args):
        return [self._convert_input(a, t) for a, t in zip(args, self.input_arguments)]

    def _convert_outputs(self, args):
        return [self._convert_output(a, t) for a, t in zip(args, self.output_arguments)]

    def __call__(self, function, funcname=None):
        if funcname is None:
            funcname = function.__name__

        def _wrapped_function(*args, **kwdargs):
            bound = check_mode(args, list(self._extract_callmode()), funcname, **kwdargs)
            converted_args = self._convert_inputs(args)
            result = function(*converted_args)
            if len(self.output_arguments) == 1:
                result = [result]

            try:
                transformed = []
                for i, r in enumerate(result):
                    r = self._convert_output(r, self.output_arguments[i])
                    if bound & (1 << (len(self.output_arguments) - i - 1)):
                        r = unify_value(r, args[len(self.input_arguments) + i], {})
                    transformed.append(r)
                result = args[:len(self.input_arguments)] + tuple(transformed)
                return [result]
            except UnifyError:
                return []

        problog_export.add_function(funcname, len(self.input_arguments),
                                    len(self.output_arguments), _wrapped_function)
        return function


# noinspection PyPep8Naming
class problog_export_raw(problog_export):

    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwdargs):
        problog_export.__init__(self, *args, **kwdargs)
        self.input_arguments = [a[1:] for a in args]

    def _convert_input(self, a, t):
        if is_variable(a):
            return None
        else:
            return problog_export._convert_input(self, a, t)

    def _extract_callmode(self):
        callmode_in = ''

        # multiple call modes: index = binary encoding on whether the output is bound
        # 0 -> all unbound
        # 1 -> first output arg is bound
        # 2 -> second output arg is bound
        # 3 -> first and second are bound

        n = len(self.input_arguments)
        for i in range(0, 1 << n):
            callmode = callmode_in
            for j, t in enumerate(self.input_arguments):
                if i & (1 << (n - j - 1)):
                    callmode += self._type_to_callmode(t)
                else:
                    callmode += 'v'
            yield callmode

    def __call__(self, function, funcname=None):
        if funcname is None:
            funcname = function.__name__


        def _wrapped_function(*args, **kwdargs):
            bound = check_mode(args, list(self._extract_callmode()), funcname, **kwdargs)
            converted_args = self._convert_inputs(args)
            results = []
            for result in function(*converted_args):
                if len(result) == 2 and type(result[0]) == tuple:
                    # Probabilistic
                    result, p = result
                    raise Exception('We don\'t support probabilistic yet!')
                else:
                    p = None

                # result is always a list of tuples
                try:
                    transformed = []
                    for i, r in enumerate(result):
                        r = self._convert_output(r, self.input_arguments[i])
                        if bound & (1 << i):
                            r = unify_value(r, args[i], {})
                        transformed.append(r)
                    result = tuple(transformed)
                    results.append(result)
                except UnifyError:
                    pass
            return results

        problog_export.add_function(funcname, len(self.input_arguments),
                                    0, _wrapped_function)
        return function


# noinspection PyPep8Naming
class problog_export_nondet(problog_export):
    def __call__(self, function, funcname=None):
        if funcname is None:
            funcname = function.__name__

        def _wrapped_function(*args, **kwdargs):
            bound = check_mode(args, list(self._extract_callmode()), funcname, **kwdargs)
            converted_args = self._convert_inputs(args)
            results = []
            for result in function(*converted_args):
                if len(self.output_arguments) == 1:
                    result = [result]

                try:
                    transformed = []
                    for i, r in enumerate(result):
                        r = self._convert_output(r, self.output_arguments[i])
                        if bound & (1 << (len(self.output_arguments) - i - 1)):
                            r = unify_value(r, args[len(self.input_arguments) + i], {})
                        transformed.append(r)
                    result = args[:len(self.input_arguments)] + tuple(transformed)
                    results.append(result)
                except UnifyError:
                    pass
            return results

        problog_export.add_function(funcname, len(self.input_arguments),
                                    len(self.output_arguments), _wrapped_function)
        return function


def _builtin_use_module(filename, database=None, location=None, **kwdargs):
    if filename.functor == 'library' and filename.arity == 1:
        filename = os.path.join(os.path.dirname(__file__), 'library',
                                _atom_to_filename(filename.args[0]))
        if not os.path.exists(filename + '.pl'):
            filename += '.py'
    else:
        root = database.source_root
        if filename.location:
            source_root = database.source_files[filename.location[0]]
            if source_root:
                root = os.path.dirname(source_root)

        filename = os.path.join(root, _atom_to_filename(filename))

    if filename[-3:] == '.py':
        try:
            load_external_module(database, filename)
        except IOError as err:
            raise ConsultError('Error while reading external library: %s' % str(err),
                               database.lineno(location))
        return True
    else:
        return _builtin_consult(Term(filename), database=database, location=location, **kwdargs)


def load_external_module(database, filename):
    import imp
    problog_export.database = database
    with open(filename, 'r') as extfile:
        imp.load_module('externals', extfile, filename, ('.py', 'U', 1))


def _builtin_call(term, args=(), engine=None, callback=None, transform=None, context=None, **kwdargs):
    check_mode((term,), ['c'], functor='call')
    # Find the define node for the given query term.
    term_call = term.with_args(*(term.args + args))

    try:
        if transform is None:
            from .engine_stack import Transformations
            transform = Transformations()

        def _trans(result):
            n = len(term.args)
            res1 = result[:n]
            res2 = result[n:]
            return [term.with_args(*res1)] + list(res2)
        transform.addFunction(_trans)

        actions = engine.call_intern(term_call, transform=transform, **kwdargs)
    except UnknownClauseInternal:
        raise UnknownClause(term_call.signature, kwdargs['database'].lineno(kwdargs['location']))
    return True, actions


def _builtin_subquery(term, prob, evidence=None, engine=None, database=None, **kwdargs):
    if evidence:
        check_mode((term, prob, evidence), ['cvL'], functor='subquery')
    else:
        check_mode((term, prob), ['cv'], functor='subquery')

    from .sdd_formula import SDD

    eng = engine.__class__()

    target = eng.ground(database, term, label='query')

    if evidence:
        for ev in term2list(evidence):
            target = eng.ground(database, ev, target=target, label=target.LABEL_EVIDENCE_POS)

    results = SDD.create_from(target).evaluate()
    if evidence:
        return [(t, Constant(p), evidence) for t, p in results.items()]
    else:
        return [(t, Constant(p)) for t, p in results.items()]


def _builtin_calln(term, *args, **kwdargs):
    return _builtin_call(term, args, **kwdargs)


class IndirectCallCycleError(GroundingError):
    """Cycle should not pass through indirect calls (e.g. call/1, findall/3)."""

    def __init__(self, location=None):
        GroundingError.__init__(self,
                                'Indirect cycle detected (passing through call/1 or findall/3)',
                                location)



"""
problog.setup - Installation tools
----------------------------------

Provides an installer for ProbLog dependencies.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

import os
import sys
import distutils.spawn
import distutils.ccompiler


def get_system():
    system = sys.platform
    if system.lower().startswith('java'):
        import java.lang.System
        system = java.lang.System.getProperty('os.name').lower()
    if system.startswith('linux'):
        system = 'linux'
    elif system.startswith('win'):
        system = 'windows'
    elif system.startswith('mac'):
        system = 'darwin'
    return system


def set_environment():
    """Updates local PATH and PYTHONPATH to include additional component directories."""
    # Set PATH environment
    path = os.environ.get('PATH', [])
    if path:
        path = [path]
    path = os.pathsep.join(path + list(get_binary_paths()))
    os.environ['PATH'] = path
    # Set PYTHONPATH environment
    sys.path += get_module_paths()


def get_binary_paths():
    """Get a list of additional binary search paths."""
    binary_root = os.path.join(os.path.dirname(__file__), 'bin')
    system = get_system()  # Darwin, Linux, ?
    return list(map(os.path.abspath, [os.path.join(binary_root, system), binary_root]))


def get_module_paths():
    """Get a list of additional module search paths."""
    binary_root = os.path.join(os.path.dirname(__file__), 'lib')
    system = get_system()  # Darwin, Linux, ?
    python = 'python%s' % sys.version_info[0]
    return list(map(os.path.abspath,
                    [os.path.join(binary_root, python, system),
                     os.path.join(binary_root, python),
                     binary_root]))


def gather_info():
    """Collect info about the system and its installed software."""

    system_info = {}

    system_info['root_path'] = os.path.join(os.path.dirname(__file__), '..')

    system_info['os'] = get_system()
    # system_info['arch'] = os.uname()[-1]

    system_info['python_version'] = sys.version_info

    # Module pyparsing
    try:
        import pyparsing
        system_info['pyparsing'] = pyparsing.__version__
    except ImportError:
        pass

    # SDD module
    # noinspection PyBroadException
    try:
        import sdd
        system_info['sdd_module'] = True
    except Exception:
        pass

    # DSharp
    system_info['dsharp'] = distutils.spawn.find_executable('dsharp') is not None

    # c2d
    system_info['c2d'] = distutils.spawn.find_executable('cnf2dDNNF') is not None
    return system_info


def build_sdd():
    if get_system() == 'windows':
        print('The SDD library is not yet available for Windows.')
        return

    build_lib = get_module_paths()[0]
    build_dir = get_module_paths()[-1]

    lib_dir = os.path.abspath(os.path.join(build_dir, 'sdd', os.uname()[0].lower()))

    curr = os.curdir
    os.chdir(build_dir)

    from distutils.core import setup, Extension
    sdd_module = Extension('_sdd', sources=['sdd/sdd_wrap.c', 'sdd/except.c'], libraries=['sdd'],
                           library_dirs=[lib_dir])

    setup(name='sdd',
          version='1.0',
          author="",
          description="""SDD Library""",
          ext_modules=[sdd_module],
          py_modules=["sdd"],
          script_name='',
          script_args=['build_ext', '--build-lib', build_lib, '--rpath', lib_dir]
          )

    os.chdir(curr)


def build_maxsatz():
    if get_system() == 'windows':
        return  # We include the binary

    compiler = distutils.ccompiler.new_compiler()

    dest_dir, source_dir = get_binary_paths()
    source_dir = os.path.join(source_dir, 'source', 'maxsatz')
    source_file = 'maxsatz2009.c'
    curdir = os.curdir
    os.chdir(source_dir)
    objfile = compiler.compile([source_file], output_dir=dest_dir)
    compiler.link_executable(objfile, os.path.join(dest_dir, 'maxsatz'))


def install(force=True):
    info = gather_info()
    update = False

    if force or not info.get('sdd_module'):
        build_sdd()
        update = True

    # build_maxsatz()

    if update:
        info = gather_info()
    return info


def system_info():
    info = gather_info()

    ok = True
    s = 'System information:\n'
    s += '------------------:\n'
    s += 'Operating system: %s\n' % info.get('os', 'unknown')
    s += 'System architecture: %s\n' % info.get('arch', 'unknown')
    s += 'Python version: %s.%s.%s\n' % (
        info['python_version'].major, info['python_version'].minor, info['python_version'].micro)
    s += '\n'
    s += 'ProbLog components:\n'
    s += '-------------------\n'

    # PrologFile, PrologString  => require pyparsing
    # SDD => requires sdd_library
    # NNF => requires dsharp or c2d

    # SemiringOther => requires NNF (or SDD with alternative evaluation)

    # pyparsing = info.get('pyparsing', 'NOT INSTALLED')
    # s += 'Module \'pyparsing\': %s\n' % pyparsing
    # if not pyparsing :
    #     s += '  ACTION: install the pyparsing module\n'
    # sdd = info.get('sdd_module', False)
    # if sdd :
    #     s += 'Module \'sdd\': INSTALLED\n'
    # else :
    #     s += 'Module \'sdd\': NOT INSTALLED\n'
    #     s += '  ACTION: run ProbLog installer\n'
    #
    return s


if __name__ == '__main__':
    set_environment()
    info = install()
    print(info)



"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function

import os

def build_sdd():

    build_lib = '.'  # where to put final library?
    build_dir = '.'  # where to put temporary build files?

    lib_dir = '.'   # where to find SDD library

    curr = os.curdir
    os.chdir(build_dir)

    from distutils.core import setup, Extension
    sdd_module = Extension('_sdd', sources=['sdd_wrap.c', 'except.c'], libraries=['sdd'], library_dirs=[lib_dir])

    setup (name='sdd',
           version='1.0',
           author="",
           description="""SDD Library""",
           ext_modules=[sdd_module],
           py_modules=["sdd"],
           script_name='',
           script_args=['build_ext', '--build-lib', build_lib, '--rpath', lib_dir]
    )

    os.chdir(curr)

if __name__ == '__main__':
    build_sdd()



"""
problog.formula - Ground programs
---------------------------------

Data structures for propositional logic.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

from collections import namedtuple, defaultdict, OrderedDict

from .core import ProbLogObject
from .errors import InconsistentEvidenceError

from .util import OrderedSet
from .logic import Term, Or, Clause, And, is_ground

from .constraint import ConstraintAD


class BaseFormula(ProbLogObject):
    """Defines a basic logic formula consisting of nodes in some logical relation.

        Each node is represented by a key. This key has the following properties:
         - None indicates false
         - 0 indicates true
         - a number larger than 0 indicates a positive node
         - the key -a with a a number larger than 0 indicates the negation of a

        This data structure also support weights on nodes, names on nodes and constraints.
    """

    # Define special keys
    TRUE = 0
    FALSE = None

    WEIGHT_NEUTRAL = True
    WEIGHT_NO = None

    LABEL_QUERY = "query"
    LABEL_EVIDENCE_POS = "evidence+"
    LABEL_EVIDENCE_NEG = "evidence-"
    LABEL_EVIDENCE_MAYBE = "evidence?"
    LABEL_NAMED = "named"

    def __init__(self):
        self._weights = {}               # Node weights: dict(key: Term)

        self._constraints = []           # Constraints: list of Constraint

        self._names = defaultdict(OrderedDict)  # Node names: dict(label: dict(key, Term))

        self._atomcount = 0

    @property
    def atomcount(self):
        """Number of atoms in the formula."""
        return self._atomcount

    # ====================================================================================== #
    # ==========                          NODE WEIGHTS                           =========== #
    # ====================================================================================== #

    def get_weights(self):
        """Get weights of the atoms in the formula.

        :return: dictionary of weights
        :rtype: dict[int, Term]
        """
        return self._weights

    def set_weights(self, weights):
        """Set weights of the atoms in the formula.

        :param weights: dictionary of weights
        :type weights: dict[int, Term]
        """
        self._weights = weights

    def get_weight(self, key, semiring):
        """Get actual value of the node with the given key according to the given semiring.

        :param key: key of the node (can be TRUE, FALSE or positive or negative)
        :param semiring: semiring to use to transform stored weight term into actual value
        :type semiring: problog.evaluator.Semiring
        :return: actual value of the weight of the given node
        """
        if self.is_false(key):
            return semiring.zero()
        elif self.is_true(key):
            return semiring.one()
        elif key < 0:
            return semiring.neg_value(self._weights[-key], key)
        else:
            return semiring.pos_value(self._weights[key], key)

    def extract_weights(self, semiring, weights=None):
        """Extracts the positive and negative weights for all atoms in the data structure.

        :param semiring: semiring that determines the interpretation of the weights
        :param weights: dictionary of { node name : weight } that overrides the builtin weights
        :returns: dictionary { key: (positive weight, negative weight) }
        :rtype: dict[int, tuple[any]]

        Atoms with weight set to neutral will get weight ``(semiring.one(), semiring.one())``.

        If the weights argument is given, it completely replaces the formula's weights.

        All constraints are applied to the weights.
        """

        if weights is None:
            weights = self.get_weights()
        else:
            oweights = dict(self.get_weights().items())
            oweights.update({self.get_node_by_name(n): v for n, v in weights.items()})
            weights = oweights

        result = {}
        for n, w in weights.items():
            if w == self.WEIGHT_NEUTRAL and type(self.WEIGHT_NEUTRAL) == type(w):
                result[n] = semiring.one(), semiring.one()
            else:
                result[n] = semiring.pos_value(w, n), semiring.neg_value(w, n)

        for c in self.constraints():
            c.update_weights(result, semiring)
        return result

    # ====================================================================================== #
    # ==========                           NODE NAMES                            =========== #
    # ====================================================================================== #

    def add_name(self, name, key, label=None):
        """Add a name to the given node.

        :param name: name of the node
        :type name: Term
        :param key: key of the node
        :type key: int | TRUE | FALSE
        :param label: type of label (one of LABEL_*)
        """
        if label is None:
            label = self.LABEL_NAMED
        self._names[label][name] = key

    def get_node_by_name(self, name):
        """Get node corresponding to the given name.

        :param name: name of the node to find
        :return: key of the node
        :raises: KeyError if no node with the given name was found
        """
        for names in self._names.values():
            res = names.get(name, '#NOTFOUND#')
            if res != '#NOTFOUND#':
                return res
        raise KeyError(name)

    def add_query(self, name, key):
        """Add a query name.

        Same as ``add_name(name, key, self.LABEL_QUERY)``.

        :param name: name of the query
        :param key: key of the query node
        """
        self.add_name(name, key, self.LABEL_QUERY)

    def add_evidence(self, name, key, value):
        """Add an evidence name.

        Same as ``add_name(name, key, self.LABEL_EVIDENCE_???)``.

        :param name: name of the query
        :param key: key of the query node
        :param value: value of the evidence (True, False or None)
        """
        if value is None:
            self.add_name(name, key, self.LABEL_EVIDENCE_MAYBE)
        elif value:
            self.add_name(name, key, self.LABEL_EVIDENCE_POS)
        else:
            self.add_name(name, key, self.LABEL_EVIDENCE_NEG)

    def get_names(self, label=None):
        """Get a list of all node names in the formula.

        :param label: restrict to given label. If not set, all nodes are returned.
        :return: list of all nodes names (of the requested type) as a list of tuples (name, key)
        """
        if label is None:
            result = OrderedSet()
            for names in self._names.values():
                for name, node in names.items():
                    result.add((name, node))
            return result
        else:
            return self._names.get(label, {}).items()

    def get_names_with_label(self):
        """Get a list of all node names in the formula with their label type.

        :return: list of all nodes names with their type
        """
        result = []
        for label in self._names:
            for name, key in self._names[label].items():
                result.append((name, key, label))
        return result

    def queries(self):
        """Get a list of all queries.

        :return: ``get_names(LABEL_QUERY)``
        """
        return self.get_names(self.LABEL_QUERY)

    def evidence(self):
        """Get a list of all determined evidence.
        Keys are negated for negative evidence.
        Unspecified evidence (value None) is not included.

        :return: list of tuples (name, key) for positive and negative evidence
        """
        evidence_true = self.get_names(self.LABEL_EVIDENCE_POS)
        evidence_false = self.get_names(self.LABEL_EVIDENCE_NEG)
        return list(evidence_true) + [(name, self.negate(node)) for name, node in evidence_false]

    def evidence_all(self):
        """Get a list of all evidence (including undetermined).

        :return: list of tuples (name, key, value) where value can be -1, 0 or 1
        """
        evidence_true = [x + (1,) for x in self.get_names(self.LABEL_EVIDENCE_POS)]
        evidence_false = [x + (-1,) for x in self.get_names(self.LABEL_EVIDENCE_NEG)]
        evidence_maybe = [x + (0,) for x in self.get_names(self.LABEL_EVIDENCE_MAYBE)]
        return evidence_true + evidence_false + evidence_maybe

    # ====================================================================================== #
    # ==========                        KEY MANIPULATION                         =========== #
    # ====================================================================================== #

    def is_true(self, key):
        """Does the key represent deterministic True?

        :param key: key
        :return: ``key == self.TRUE``
        """
        return key == self.TRUE

    def is_false(self, key):
        """Does the key represent deterministic False?

        :param key: key
        :return: ``key == self.FALSE``
        """

        return key == self.FALSE

    def is_probabilistic(self, key):
        """Does the key represent a probabilistic node?

        :param key: key
        :return: ``not is_true(key) and not is_false(key)``
        """
        return not self.is_true(key) and not self.is_false(key)

    def negate(self, key):
        """Negate the key.

        For TRUE, returns FALSE;
        For FALSE, returns TRUE;
        For x returns -x

        :param key: key to negate
        :return: negation of the key
        """
        if key == self.TRUE:
            return self.FALSE
        elif key == self.FALSE:
            return self.TRUE
        else:
            return -key

    # ====================================================================================== #
    # ==========                          CONSTRAINTS                            =========== #
    # ====================================================================================== #

    def constraints(self):
        """Return the list of constraints.

        :return: list of constraints
        """
        return self._constraints

    def add_constraint(self, constraint):
        """Add a constraint

        :param constraint: constraint to add
        :type constraint: problog.constraint.Constraint
        """
        self._constraints.append(constraint)

    def flag(self, flag):
        flag = '_%s' % flag
        return hasattr(self, flag) and getattr(self, flag)


class LogicFormula(BaseFormula):
    """A logic formula is a data structure that is used to represent generic And-Or graphs.
    It can typically contain three types of nodes:

        - atom ( or terminal)
        - and (compound)
        - or (compound)

    The compound nodes contain a list of children which point to other nodes in the formula.
    These pointers can be positive or negative.

    In addition to the basic logical structure of the formula, it also maintains a table of labels,
    which can be used to easily retrieve certain nodes.
    These labels typically contain the literals from the original program.

    Upon addition of new nodes, the logic formula can perform certain optimizations, for example,
    by simplifying nodes or by reusing existing nodes.
    """

    _atom = namedtuple('atom', ('identifier', 'probability', 'group', 'name', 'source'))
    _conj = namedtuple('conj', ('children', 'name'))
    _disj = namedtuple('disj', ('children', 'name'))
    # negation is encoded by using a negative number for the key

    def _create_atom(self, identifier, probability, group, name=None, source=None):
        return self._atom(identifier, probability, group, name, source)

    def _create_conj(self, children, name=None):
        return self._conj(children, name)

    def _create_disj(self, children, name=None):
        return self._disj(children, name)

    # noinspection PyUnusedLocal
    def __init__(self, auto_compact=True, avoid_name_clash=False, keep_order=False,
                 use_string_names=False, keep_all=False, propagate_weights=None,
                 max_arity=0, keep_duplicates=False, keep_builtins=False, hide_builtins=False,
                 **kwdargs):
        BaseFormula.__init__(self)

        # List of nodes
        self._nodes = []
        # Lookup index for 'atom' nodes, key is identifier passed to addAtom()
        self._index_atom = {}
        # Lookup index for 'and' nodes, key is tuple of sorted children
        self._index_conj = {}
        # Lookup index for 'or' nodes, key is tuple of sorted children
        self._index_disj = {}

        self._atomcount = 0

        self._auto_compact = auto_compact
        self._avoid_name_clash = avoid_name_clash
        self._keep_order = keep_order
        self._keep_all = keep_all
        self._keep_builtins = (keep_all or keep_builtins) and not hide_builtins
        self._keep_duplicates = keep_duplicates

        self._max_arity = max_arity

        self._constraints_me = {}

        self.semiring = propagate_weights

        self._use_string_names = use_string_names

    # ====================================================================================== #
    # ==========                         MANAGE LABELS                           =========== #
    # ====================================================================================== #

    def add_name(self, name, key, label=None):
        """Associates a name to the given node identifier.

            :param name: name of the node
            :param key: id of the node
            :param label: type of node (see LogicFormula.LABEL_*)
        """
        if self._use_string_names:
            name = str(name)

        if self.is_probabilistic(key):
            node = self.get_node(abs(key))
            ntype = type(node).__name__
            if key < 0:
                lname = -name
            else:
                lname = name
            if ntype == 'atom':
                node = type(node)(*(node[:-2] + (lname, node[-1])))
            else:
                node = type(node)(*(node[:-1] + (lname,)))
            self._update(abs(key), node)

        BaseFormula.add_name(self, name, key, label)

    # ====================================================================================== #
    # ==========                          MANAGE LOGIC                           =========== #
    # ====================================================================================== #

    def is_trivial(self):
        """Test whether the formula contains any logical construct.

        :return: False if the formula only contains atoms.
        """
        return self.atomcount == len(self)

    # noinspection PyUnusedLocal
    def _add(self, node, key=None, reuse=True):
        """Adds a new node, or reuses an existing one.

        :param node: node to add
        :param reuse: (default True) attempt to map the new node onto an existing one based on its \
         content

        """
        if reuse:
            # Determine the node's key and lookup identifier base on node type.
            ntype = type(node).__name__
            if ntype == 'atom':
                key = node.identifier
                collection = self._index_atom
            elif ntype == 'conj':
                key = node.children
                collection = self._index_conj
            elif ntype == 'disj':
                key = node.children
                collection = self._index_disj
            else:
                raise TypeError("Unexpected node type: '%s'." % ntype)

            if key not in collection:
                # Create a new entry, starting from 1
                index = len(self._nodes) + 1
                # Add the entry to the collection
                collection[key] = index
                # Add entry to the set of nodes
                self._nodes.append(node)
            else:
                # Retrieve the entry from collection
                index = collection[key]
        else:
            # Don't reuse, just add node.
            index = len(self._nodes) + 1
            self._nodes.append(node)

        # Return the entry
        return index

    def _update(self, key, value):
        """Replace the node with the given content.

        :param key: key of the node to replace.
        :type key: int > 0
        :param value: new content of the node
        """
        assert(self.is_probabilistic(key))
        assert(key > 0)
        self._nodes[key - 1] = value

    def _add_constraint_me(self, group, node):
        if group is None:
            return node
        constraint = self._constraints_me.get(group)
        if constraint is None:
            constraint = ConstraintAD(group)
            self._constraints_me[group] = constraint
        node = constraint.add(node, self)
        return node

    def add_atom(self, identifier, probability, group=None, name=None, source=None):
        """Add an atom to the formula.

        :param identifier: a unique identifier for the atom
        :param probability: probability of the atom
        :param group: a group identifier that identifies mutually exclusive atoms (or None if no \
        constraint)
        :param name: name of the new node
        :returns: the identifiers of the node in the formula (returns self.TRUE for deterministic \
        atoms)

        This function has the following behavior :

        * If ``probability`` is set to ``None`` then the node is considered to be deterministically\
         true and the function will return :attr:`TRUE`.
        * If a node already exists with the given ``identifier``, the id of that node is returned.
        * If ``group`` is given, a mutual exclusivity constraint is added for all nodes sharing the\
         same group.
        * To add an explicitly present deterministic node you can set the probability to ``True``.
        """
        if probability is None and not self._keep_all:
            return self.TRUE
        elif probability != self.WEIGHT_NEUTRAL and self.semiring and \
                self.semiring.is_zero(self.semiring.value(probability)):
            return self.FALSE
        elif probability != self.WEIGHT_NEUTRAL and self.semiring and \
                self.semiring.is_one(self.semiring.value(probability)):
            return self.TRUE
        else:
            atom = self._create_atom(identifier, probability, group, name, source)
            node_id = self._add(atom, key=identifier)

            self.get_weights()[node_id] = probability
            if name is not None:
                self.add_name(name, node_id, self.LABEL_NAMED)
            if node_id == len(self._nodes):
                # The node was not reused?
                self._atomcount += 1
                # TODO if the next call return 0 or None, the node is still added?
                node_id = self._add_constraint_me(group, node_id)
            return node_id

    def add_and(self, components, key=None, name=None):
        """Add a conjunction to the logic formula.

        :param components: a list of node identifiers that already exist in the logic formula.
        :param key: preferred key to use
        :param name: name of the node
        :returns: the key of the node in the formula (returns 0 for deterministic atoms)
        """
        return self._add_compound('conj', components, self.FALSE, self.TRUE, key=key, name=name)

    def add_or(self, components, key=None, readonly=True, name=None):
        """Add a disjunction to the logic formula.

        :param components: a list of node identifiers that already exist in the logic formula.
        :param key: preferred key to use
        :param readonly: indicates whether the node should be modifiable. This will allow \
        additional disjunct to be added without changing the node key. Modifiable nodes are \
        less optimizable.
        :param name: name of the node
        :returns: the key of the node in the formula (returns 0 for deterministic atoms)
        :rtype: :class:`int`

        By default, all nodes in the data structure are immutable (i.e. readonly).
        This allows the data structure to optimize nodes, but it also means that cyclic formula \
        can not be stored because the identifiers of all descendants must be known add creation \
        time.

        By setting `readonly` to False, the node is made mutable and will allow adding disjunct \
        later using the :func:`addDisjunct` method.
        This may cause the data structure to contain superfluous nodes.
        """
        return self._add_compound('disj', components, self.TRUE, self.FALSE, key=key,
                                  readonly=readonly, name=name)

    def add_disjunct(self, key, component):
        """Add a component to the node with the given key.

        :param key: id of the node to update
        :param component: the component to add
        :return: key
        :raises: :class:`ValueError` if ``key`` points to an invalid node

        This may only be called with a key that points to a disjunctive node or :attr:`TRUE`.
        """
        if self.is_true(key):
            return key
        elif self.is_false(key):
            raise ValueError("Cannot update failing node")
        else:
            node = self.get_node(key)
            if type(node).__name__ != 'disj':
                raise ValueError("Can only update disjunctive node")

            if component is None:
                # Don't do anything
                pass
            elif component == 0:
                return self._update(key, self._create_disj((0,), name=node.name))
            elif component in node.children and not self._keep_duplicates:
                pass    # already there
            else:
                if 0 < self._max_arity == len(node.children):
                    child = self.add_or(node.children)
                    return self._update(key, self._create_disj((child, component), name=node.name))
                else:
                    return self._update(key, self._create_disj(node.children + (component,),
                                                               name=node.name))
            return key

    def add_not(self, component):
        """Returns the key to the negation of the node.

        :param component: the node to negate
        """
        return self.negate(component)

    def get_node(self, key):
        """Get the content of the node with the given key.

        :param key: key of the node
        :type key: int > 0
        :return: content of the node
        """
        assert self.is_probabilistic(key)
        assert key > 0
        return self._nodes[key - 1]

    # noinspection PyUnusedLocal
    def _add_compound(self, nodetype, content, t, f, key=None,
                      readonly=True, update=None, name=None):
        """Add a compound term (AND or OR)."""
        assert content   # Content should not be empty

        name_clash = False
        if self._auto_compact:
            # If there is a t node, (true for OR, false for AND)
            if t in content:
                return t

            # Eliminate unneeded node nodes (false for OR, true for AND)
            content = filter(lambda x: x != f, content)

            # Put into fixed order and eliminate duplicate nodes
            if self._keep_duplicates:
                content = tuple(content)
            elif self._keep_order:
                content = tuple(OrderedSet(content))
            else:  # any_order
                # can also merge (a, b) and (b, a)
                content = tuple(set(content))

            # Empty OR node fails, AND node is true
            if not content:
                return f

            # Contains opposites: return 'TRUE' for or, 'FALSE' for and
            if len(set(content)) > len(set(map(abs, content))):
                return t

            # If node has only one child, just return the child.
            # Don't do this for modifiable nodes, we need to keep a separate node.
            if (readonly and update is None) and len(content) == 1:
                if self._avoid_name_clash:
                    name_old = self.get_node(abs(content[0])).name
                    if name is None or name_old is None or name == name_old:
                        if name is not None:
                            self.add_name(name, content[0], self.LABEL_NAMED)
                        return content[0]
                    else:
                        name_clash = True
                else:
                    if name is not None:
                        name_old = self.get_node(abs(content[0])).name
                        if name_old is None:
                            self.add_name(name, content[0], self.LABEL_NAMED)

                    return content[0]
        else:
            content = tuple(content)

        if nodetype == 'conj':
            node = self._create_conj(content, name)
            return self._add(node, reuse=self._auto_compact and not self._keep_all)
        elif nodetype == 'disj':
            node = self._create_disj(content, name)
            if update is not None:
                # If an update key is set, update that node
                return self._update(update, node)
            elif readonly:
                # If the node is readonly, we can try to reuse an existing node.
                new_node = self._add(node, reuse=self._auto_compact and not name_clash and not self._keep_all)
                return new_node
            else:
                # If node is modifiable, we shouldn't reuse an existing node.
                return self._add(node, reuse=False)

        else:
            raise TypeError("Unexpected node type: '%s'." % nodetype)

    def __iter__(self):
        """Iterate over the nodes in the formula.

        :returns: iterator over tuples ( key, node, type )
        """
        for i, n in enumerate(self._nodes):
            yield (i + 1, n, type(n).__name__)

    def __len__(self):
        """Returns the number of nodes in the formula."""
        return len(self._nodes)

    # ====================================================================================== #
    # ==========                       MANAGE CONSTRAINTS                        =========== #
    # ====================================================================================== #

    def constraints(self):
        """Returns a list of all constraints."""
        return list(self._constraints_me.values()) + BaseFormula.constraints(self)

    # ====================================================================================== #
    # ==========                      EVIDENCE PROPAGATION                       =========== #
    # ====================================================================================== #

    def has_evidence_values(self):
        """Checks whether the current formula contains information for evidence propagation."""
        return hasattr(self, 'lookup_evidence')

    def get_evidence_values(self):
        """Retrieves evidence propagation information."""
        assert(self.has_evidence_values())
        return getattr(self, 'lookup_evidence')

    def get_evidence_value(self, key):
        """Get value of the given node based on evidence propagation.

        :param key: key of the node
        :return: value of the node (key, TRUE or FALSE)
        """
        if key == 0 or key is None:
            return key
        elif self.has_evidence_values():
            result = self.get_evidence_values().get(abs(key), abs(key))
            if key < 0:
                return self.negate(result)
            else:
                return result
        else:
            return key

    def set_evidence_value(self, key, value):
        """Set value of the given node based on evidence propagation.

        :param key: key of the node
        :param value: value of the node
        """
        if key < 0:
            self.get_evidence_values()[-key] = self.negate(value)
        else:
            self.get_evidence_values()[key] = value

    def propagate(self, nodeids, current=None):
        """Propagate the value of the given node
          (true if node is positive, false if node is negative)
        The propagation algorithm is not complete.

        :param nodeids: evidence nodes to set (> 0 means true, < 0 means false)
        :param current: current set of nodes with deterministic value
        :return: dictionary of nodes with deterministic value
        """

        # Initialize current in case nothing is known yet.
        if current is None:
            current = {}

        # Provide easy access to values
        values = {True: self.TRUE, False: self.FALSE}

        # Reverse mapping between a node and its parents
        atoms_in_rules = defaultdict(set)

        # Queue of nodes that need to be handled.
        # INVARIANT: elements in queue have deterministic value
        # INVARIANT: elements in queue are not yet listed in current
        queue = set(nodeids)

        while queue:
            nid = queue.pop()

            # Get information about the node
            n = self.get_node(abs(nid))
            t = type(n).__name__

            if abs(nid) not in current:
                # This is the first time we process this node.
                # We should process its parents again.
                for at in atoms_in_rules[abs(nid)]:
                    if at in current:
                        # Parent has a truth value.
                        # Try to propagate parent again.
                        if current[abs(at)] == self.TRUE:
                            queue.add(abs(at))
                        else:
                            queue.add(-abs(at))

            # Record node value in current
            current[abs(nid)] = values[nid > 0]

            # Process node and propagate to children
            if t == 'atom':
                # Nothing to do.
                pass
            else:
                # Get the list of children with their actual (propagated) values.
                children = []
                for c in n.children:
                    ch = current.get(abs(c), abs(c))
                    if c < 0:
                        ch = self.negate(ch)
                    children.append(ch)

                # Handle trivial cases:
                # Node should be true, but is a conjunction with a false child
                if t == 'conj' and self.FALSE in children and nid > 0:
                    raise InconsistentEvidenceError()
                # Node should be false, but is a disjunction with a true child
                elif t == 'disj' and self.TRUE in children and nid < 0:
                    raise InconsistentEvidenceError()
                # Node should be false, and is a conjunction with a false child
                elif t == 'conj' and self.FALSE in children and nid < 0:
                    # Already satisfied, nothing else to do
                    pass
                # Node should be true and is a disjunction with a true child
                elif t == 'disj' and self.TRUE in children and nid > 0:
                    # Already satisfied, nothing else to do
                    pass
                else:
                    # Filter out deterministic children
                    children = list(filter(lambda x: x != 0 and x is not None, children))
                    if len(children) == 1:
                        # One child left: propagate value to the child
                        if abs(children[0]) not in current:
                            if nid < 0:
                                queue.add(-children[0])
                            else:
                                queue.add(children[0])
                            atoms_in_rules[abs(children[0])].discard(abs(nid))
                    elif nid > 0 and t == 'conj':
                        # Conjunction is true => all children are true
                        for c in children:
                            if abs(c) not in current:
                                queue.add(c)
                            atoms_in_rules[abs(c)].discard(abs(nid))
                    elif nid < 0 and t == 'disj':
                        # Disjunction is false => all children are false
                        for c in children:
                            if abs(c) not in current:
                                queue.add(-c)
                            atoms_in_rules[abs(c)].discard(abs(nid))
                    else:
                        # We can't propagate yet. Mark current rule as parent of its children.
                        for c in children:
                            atoms_in_rules[abs(c)].add(abs(nid))
        return current

    # def propagate(self, nodeids, current=None):
    #     if current is None:
    #         current = {}
    #
    #     values = {True: self.TRUE, False: self.FALSE}
    #     atoms_in_rules = defaultdict(set)
    #
    #     updated = set()
    #     queue = set(nodeids)
    #     while queue:
    #         nid = queue.pop()
    #
    #         if abs(nid) not in current:
    #             updated.add(abs(nid))
    #             for at in atoms_in_rules[abs(nid)]:
    #                 if at in current:
    #                     if current[abs(at)] == 0:
    #                         queue.add(abs(at))
    #                     else:
    #                         queue.add(-abs(at))
    #             current[abs(nid)] = values[nid > 0]
    #
    #         n = self.get_node(abs(nid))
    #         t = type(n).__name__
    #         if t == 'atom':
    #             pass
    #         else:
    #             children = []
    #             for c in n.children:
    #                 ch = current.get(abs(c), abs(c))
    #                 if c < 0:
    #                     ch = self.negate(ch)
    #                 children.append(ch)
    #             if t == 'conj' and None in children and nid > 0:
    #                 raise InconsistentEvidenceError()
    #             elif t == 'disj' and 0 in children and nid < 0:
    #                 raise InconsistentEvidenceError()
    #             children = list(filter(lambda x: x != 0 and x is not None, children))
    #             if len(children) == 1:  # only one child
    #                 if abs(children[0]) not in current:
    #                     if nid < 0:
    #                         queue.add(-children[0])
    #                     else:
    #                         queue.add(children[0])
    #                     atoms_in_rules[abs(children[0])].discard(abs(nid))
    #             elif nid > 0 and t == 'conj':
    #                 # Conjunction is true
    #                 for c in children:
    #                     if abs(c) not in current:
    #                         queue.add(c)
    #                     atoms_in_rules[abs(c)].discard(abs(nid))
    #             elif nid < 0 and t == 'disj':
    #                 # Disjunction is false
    #                 for c in children:
    #                     if abs(c) not in current:
    #                         queue.add(-c)
    #             else:
    #                 for c in children:
    #                     atoms_in_rules[abs(c)].add(abs(nid))
    #     return current

    # ====================================================================================== #
    # ==========                        EXPORT TO STRING                         =========== #
    # ====================================================================================== #

    def __str__(self):
        s = '\n'.join('%s: %s' % (i, n) for i, n, t in self)
        f = True
        for q in self.queries():
            if f:
                f = False
                s += '\nQueries : '
            s += '\n* %s : %s' % q

        f = True
        for q in self.evidence():
            if f:
                f = False
                s += '\nEvidence : '
            s += '\n* %s : %s' % q

        f = True
        for c in self.constraints():
            if c.is_nontrivial():
                if f:
                    f = False
                    s += '\nConstraints : '
                s += '\n* ' + str(c)
        return s + '\n'

    def to_prolog(self):
        """Convert the Logic Formula to a Prolog program.

        To make this work correctly some flags should be set on the engine and LogicFormula prior \
        to grounding.
        The following code should be used:

        .. code-block:: python

            pl = problog.program.PrologFile(input_file)
            problog.formula.LogicFormula.create_from(avoid_name_clash=True, keep_order=True, \
label_all=True)
            prologfile = gp.to_prolog()

        :return: Prolog program
        :rtype: str
        """

        lines = ['%s.' % c for c in self.enum_clauses()]

        for qn, qi in self.queries():
            if is_ground(qn):
                if qi == self.TRUE:
                    lines.append('%s.' % qn)
                elif qi == self.FALSE:
                    lines.append('%s :- fail.' % qn)
                lines.append('query(%s).' % qn)

        for qn, qi in self.evidence():
            if qi < 0:
                lines.append('evidence(%s).' % -qn)
            else:
                lines.append('evidence(%s).' % qn)

        return '\n'.join(lines)

        # lines = []
        # neg_heads = set()
        # for head, body in self.enumerate_clauses():
        #     head_name = self.get_name(head)
        #     if head_name.is_negated():
        #         pos_name = -head_name
        #         head_name = Term(pos_name.functor + '_aux', *pos_name.args)
        #         if head not in neg_heads:
        #             lines.append('%s :- %s.' % (pos_name, -head_name))
        #             neg_heads.add(head)
        #     if body:    # clause with a body
        #         body = ', '.join(map(str, map(self.get_name, body)))
        #         lines.append('%s :- %s.' % (head_name, body))
        #     else:   # fact
        #         prob = self.get_node(head).probability
        #
        #         if prob is not None:
        #             lines.append('%s::%s.' % (prob, head_name))
        #         else:
        #             lines.append('%s.' % head_name)
        # return '\n'.join(lines)

    def get_name(self, key):
        """Get the name of the given node.

        :param key: key of the node
        :return: name of the node
        :rtype: Term
        """
        if key == 0:
            return Term('true')
        elif key is None:
            return Term('false')
        else:
            node = self.get_node(abs(key))
            name = node.name

            if not self._is_valid_name(name) and type(node).__name__ == 'disj' and node.children:
                if key < 0:
                    name = self.get_name(-node.children[0])
                else:
                    name = self.get_name(node.children[0])

            if name is None:
                name = Term('node_%s' % abs(key))
            if key < 0:
                return -name
            else:
                return name

    def enumerate_clauses(self, relevant_only=True):
        """Enumerate the clauses of this logic formula.
            Clauses are represented as (``head, [body]``).

        :param relevant_only: only list clauses that are part of the ground program for a query or \
         evidence
        :return: iterator of clauses
        """
        enumerated = OrderedSet()

        if relevant_only:
            to_enumerate = OrderedSet(abs(n) for q, n in self.queries() if self.is_probabilistic(n))
            to_enumerate |= OrderedSet(abs(n)
                                       for q, n in self.evidence() if self.is_probabilistic(n))
        else:
            to_enumerate = OrderedSet(range(1, len(self) + 1))

        while to_enumerate:
            i = to_enumerate.pop()
            enumerated.add(i)
            n = self.get_node(i)
            t = type(n).__name__

            if t == 'atom':
                yield i, []
            elif t == 'conj':
                # In case a query or evidence is directly referring to a conjunctive node.
                body = self._unroll_conj(n)
                to_enumerate |= (OrderedSet((map(abs, body))) - enumerated)
                yield i, body
            else:   # t == 'disj'
                for c_i in n.children:
                    negc = (c_i < 0)
                    c_n = self.get_node(abs(c_i))
                    c_t = type(c_n).__name__
                    if c_t == 'atom':
                        yield i, [c_i]
                        if abs(c_i) not in enumerated:
                            to_enumerate.add(abs(c_i))
                    elif c_t == 'conj':
                        if negc:
                            yield i, [c_i]
                            if abs(c_i) not in enumerated:
                                to_enumerate.add(abs(c_i))
                        else:
                            body = self._unroll_conj(c_n)
                            to_enumerate |= (OrderedSet(map(abs, body)) - enumerated)
                            yield i, body
                    else:
                        yield i, [c_i]
                        if abs(c_i) not in enumerated:
                            to_enumerate.add(abs(c_i))

    def extract_ads(self, relevant, processed):

        # Collect information about annotated disjunctions
        choices = set([])
        choice_group = {}
        choice_by_group = defaultdict(list)
        choice_parent = {}
        choice_by_parent = {}
        choice_body = {}
        choice_name = {}
        choice_prob = {}

        for i, n, t in self:
            if not relevant[i]:
                continue
            if t == 'atom' and n.group is not None:
                choice_group[i] = n.group
                choice_prob[i] = n.probability
                choice_by_group[n.group].append(i)
                choices.add(i)
                processed[i] = True
            elif t == 'conj' and n.children[-1] in choices:
                choice = n.children[-1]
                choice_parent[choice_group[choice]] = i
                choice_by_parent[i] = choice
                choice_body[choice_group[choice]] = n.children[0]
                if n.name is not None:
                    choice_name[choice] = n.name
                processed[i] = True
                if not self._is_valid_name(self.get_node(abs(n.children[0])).name):
                    processed[n.children[0]] = True

        for i, n, t in self:
            if t == 'disj':
                overlap = set(n.children) & set(choice_by_parent.keys()) | set(n.children) & set(choices)
                for o in overlap:
                    p = choice_by_parent.get(o, o)
                    if self._is_valid_name(n.name) and not self._is_valid_name(choice_name.get(p)):
                        choice_name[p] = n.name

        for group, choices in choice_by_group.items():
            # Construct head

            head = Or.from_list([choice_name[c].with_probability(choice_prob[c]) for c in choices])

            # Construct body
            body = self.get_body(choice_body.get(group, self.TRUE), processed)

            if body is None:
                yield head
            else:
                yield (Clause(head, body))

    def _is_valid_name(self, name):
        return name is not None and \
            not name.functor.startswith('_problog_') and \
            not name.functor == 'choice' and not name.functor == 'body'

    def get_body(self, index, processed=None, parent_name=None):
        if index == self.TRUE:
            return None
        else:
            node = self.get_node(abs(index))
            ntype = type(node).__name__
            if self._is_valid_name(node.name) and str(node.name) != str(parent_name):
                if index < 0:
                    return -node.name
                else:
                    return node.name
            elif ntype == 'atom':
                # Easy case: atom
                if index < 0:
                    return -node.name
                else:
                    return node.name
            elif ntype == 'conj':
                if index < 0:
                    return -node.name
                else:
                    children = self._unroll_conj(node)
                    return And.from_list(list(map(self.get_name, children)))
            elif ntype == 'disj' and len(node.children) == 1 and not self._is_valid_name(node.name):
                if processed:
                    processed[abs(index)] = True
                if index < 0:
                    b = self.get_body(-node.children[0], parent_name=parent_name)
                else:
                    b = self.get_body(node.children[0], parent_name=parent_name)
                return b
            elif ntype == 'disj':
                if index < 0:
                    return -node.name
                else:
                    return node.name
            else:
                print (self)
                print (index, node)
                raise Exception('Unexpected')

    def enum_clauses(self):
        relevant = self.extract_relevant()
        processed = [False] * (len(self) + 1)
        for ad in self.extract_ads(relevant, processed):
            yield ad
        for i, n, t in self:
            if relevant[i] and not processed[i]:
                if t == 'atom':
                    if n.name is not None and n.source not in ('builtin', 'negation'):
                        yield n.name.with_probability(n.probability)
                elif t == 'disj':
                    for c in n.children:
                        if not processed[abs(c)] or self._is_valid_name(self.get_node(abs(c)).name):
                            b = self.get_body(c, parent_name=n.name)
                            if str(n.name) != str(b):   # TODO bit of a hack?
                                yield Clause(n.name, b)
                elif t == 'conj' and n.name is None:
                    pass
                else:
                    yield Clause(n.name, self.get_body(i, parent_name=n.name))

    def extract_relevant(self):
        relevant = [False] * (len(self)+1)
        roots = set(abs(n) for q, n in self.queries() if self.is_probabilistic(n))
        roots |= set(abs(n) for q, n in self.evidence() if self.is_probabilistic(n))
        while roots:
            root = roots.pop()
            if not relevant[root]:
                relevant[root] = True
                node = self.get_node(root)
                ntype = type(node).__name__
                if ntype != 'atom':
                    for c in node.children:
                        if not relevant[abs(c)]:
                            roots.add(abs(c))
        return relevant

    def _unroll_conj(self, node):
        assert type(node).__name__ == 'conj'

        if len(node.children) == 1:
            return node.children
        elif len(node.children) == 2 and node.children[1] > 0:
            children = [node.children[0]]
            current = node.children[1]
            current_node = self.get_node(current)
            while type(current_node).__name__ == 'conj' and len(current_node.children) == 2 and \
                    not self._is_valid_name(current_node.name):
                children.append(current_node.children[0])
                current = current_node.children[1]
                if current > 0:
                    current_node = self.get_node(current)
                else:
                    current_node = None
            children.append(current)
        else:
            children = node.children
        return children

    def to_dot(self, not_as_node=True):
        """Write out in GraphViz (dot) format.

        :param not_as_node: represent negation as a node
        :return: string containing dot representation
        """

        not_as_edge = not not_as_node

        # Keep track of mutually disjunctive nodes.
        clusters = defaultdict(list)

        queries = set([(name, node) for name, node, label in self.get_names_with_label()])
        for i, n, t in self:
            if n.name is not None:
                queries.add((n.name, i))

        # Keep a list of introduced not nodes to prevent duplicates.
        negative = set([])

        s = 'digraph GP {\n'
        for index, node, nodetype in self:

            if nodetype == 'conj':
                s += '%s [label="AND", shape="box", style="filled", fillcolor="white"];\n' % index
                for c in node.children:
                    opt = ''
                    if c < 0 and c not in negative and not_as_node:
                        s += '%s [label="NOT"];\n' % c
                        s += '%s -> %s;\n' % (c, -c)
                        negative.add(c)

                    if c < 0 and not_as_edge:
                        opt = '[arrowhead="odotnormal"]'
                        c = -c
                    if c != 0:
                        s += '%s -> %s%s;\n' % (index, c, opt)
            elif nodetype == 'disj':
                s += '%s [label="OR", shape="diamond", style="filled", fillcolor="white"];\n' \
                     % index
                for c in node.children:
                    opt = ''
                    if c < 0 and c not in negative and not_as_node:
                        s += '%s [label="NOT"];\n' % c
                        s += '%s -> %s;\n' % (c, -c)
                        negative.add(c)
                    if c < 0 and not_as_edge:
                        opt = '[arrowhead="odotnormal"]'
                        c = -c
                    if c != 0:
                        s += '%s -> %s%s;\n' % (index, c, opt)
            elif nodetype == 'atom':
                if node.probability == self.WEIGHT_NEUTRAL:
                    pass
                elif node.group is None:
                    s += '%s [label="%s", shape="ellipse", style="filled", fillcolor="white"];\n' \
                         % (index, node.probability)
                else:
                    clusters[node.group].append('%s [ shape="ellipse", label="%s", '
                                                'style="filled", fillcolor="white" ];\n'
                                                % (index, node.probability))
            else:
                raise TypeError("Unexpected node type: '%s'" % nodetype)

        c = 0
        for cluster, text in clusters.items():
            if len(text) > 1:
                s += 'subgraph cluster_%s { style="dotted"; color="red"; \n\t%s\n }\n' \
                     % (c, '\n\t'.join(text))
            else:
                s += text[0]
            c += 1

        q = 0
        for name, index in set(queries):
            if name.is_negated():
                pos_name = -name
                name = Term(pos_name.functor + '_aux', *pos_name.args)
            opt = ''
            if index is None:
                index = 'false'
                if not_as_node:
                    s += '%s [label="NOT"];\n' % index
                    s += '%s -> %s;\n' % (index, 0)
                elif not_as_edge:
                    opt = ', arrowhead="odotnormal"'
                if 0 not in negative:
                    s += '%s [label="true"];\n' % 0
                    negative.add(0)
            elif index < 0:  # and not index in negative :
                if not_as_node:
                    s += '%s [label="NOT"];\n' % index
                    s += '%s -> %s;\n' % (index, -index)
                    negative.add(index)
                elif not_as_edge:
                    index = -index
                    opt = ', arrowhead="odotnormal"'
            elif index == 0 and index not in negative:
                s += '%s [label="true"];\n' % index
                negative.add(0)
            s += 'q_%s [ label="%s", shape="plaintext" ];\n' % (q, name)
            s += 'q_%s -> %s [style="dotted" %s];\n' % (q, index, opt)
            q += 1
        return s + '}'

    def clone(self, destination):
        source = self
        # TODO maintain a translation table
        for i, n, t in source:
            if t == 'atom':
                j = destination.add_atom(n.identifier, n.probability, n.group, source.get_name(i))
            elif t == 'conj':
                j = destination.add_and(n.children, source.get_name(i))
            elif t == 'disj':
                j = destination.add_or(n.children, source.get_name(i))
            else:
                raise TypeError('Unknown node type')
            assert i == j

        for name, node, label in source.get_names_with_label():
            destination.add_name(name, node, label)

        for c in source.constraints():
            if c.is_nontrivial():
                destination.add_constraint(c)

        return destination


class LogicDAG(LogicFormula):
    """A propositional logic formula without cycles."""

    def __init__(self, auto_compact=True, **kwdargs):
        LogicFormula.__init__(self, auto_compact, **kwdargs)


class DeterministicLogicFormula(LogicFormula):
    """A deterministic logic formula."""

    def __init__(self, **kwdargs):
        LogicFormula.__init__(self, **kwdargs)

    def add_atom(self, identifier, probability, group=None, name=None, source=None):
        return self.TRUE


#
#
# class StringKeyLogicFormula(LogicFormula) :
#     """A propositional logic formula consisting of and, or, not and atoms."""
#
#     TRUE = 'true'
#     FALSE = 'false'
#
#     def __init__(self) :
#         LogicFormula.__init__(self)
#
#         self.__nodes = defaultdict(list)
#
#         self.__constraints_me = {}
#         self.__constraints = []
#
#     def _add( self, node, key=None, reuse=True ) :
#         """Adds a new node, or reuses an existing one.
#
#         :param node: node to add
#         :param reuse: (default True) attempt to map the new node onto an existing one \
#         based on its content
#
#         """
#         self.__nodes[key].append(node)
#         return key
#
#     def _update( self, key, value ) :
#         """Replace the node with the given node."""
#         self.__nodes[ key ] = [value]
#
#     def add_not( self, component ) :
#         """Returns the key to the negation of the node."""
#         if self.isTrue(component) :
#             return self.FALSE
#         elif self.isFalse(component) :
#             return self.TRUE
#         elif component.startswith('-') :
#             return component[1:]
#         else :
#             return '-' + component
#
#     def _addCompound(self, nodetype, content, t, f, key=None, readonly=True, update=None) :
#         """Add a compound term (AND or OR)."""
#         assert( content )   # Content should not be empty
#
#         # #If there is a t node, (true for OR, false for AND)
#         # if t in content : return t
#         #
#         # # Eliminate unneeded node nodes (false for OR, true for AND)
#         # content = filter( lambda x : x != f, content )
#         #
#         # # Put into fixed order and eliminate duplicate nodes
#         # content = tuple(sorted(set(content)))
#         #
#         # # Empty OR node fails, AND node is true
#         # if not content : return f
#
#         # # Contains opposites: return 'TRUE' for or, 'FALSE' for and
#         # if len(set(content)) > len(set(map(abs,content))) : return t
#
#         # If node has only one child, just return the child.
#         # Don't do this for modifiable nodes, we need to keep a separate node.
#         if len(content) == 1 : return self._add(content[0], key=key)
#
#         content = tuple(content)
#
#         if nodetype == 'conj' :
#             node = self._create_conj( content )
#             return self._add( node, key=key )
#         elif nodetype == 'disj' :
#             node = self._create_disj( content )
#             if update != None :
#                 # If an update key is set, update that node
#                 return self._update( update, node )
#             elif readonly :
#                 # If the node is readonly, we can try to reuse an existing node.
#                 return self._add( node, key=key )
#             else :
#                 # If node is modifiable, we shouldn't reuse an existing node.
#                 return self._add( node, key=key, reuse=False )
#         else :
#             raise TypeError("Unexpected node type: '%s'." % nodetype)
#
#     def _resolve(self, key) :
#         if type(key) == str :
#             return key
#
#         res = self.__nodes.get(key,key)
#         if type(res) == list :
#             assert(len(res) == 1)
#             return res[0]
#         else :
#             return res
#
#     def _deref(self, x) :
#         c = x
#         neg = 0
#         while type(c) == str :
#             if c[0] == '-' :
#                 c = c[1:]
#                 neg += 1
#             x = c
#             if not c in self.__nodes : break
#             nn = self.__nodes[c]
#             if len(nn) == 1 :
#                 c = nn[0]
#             else :
#                 break
#         if neg % 2 == 0 :
#             return x
#         else :
#             return '-' + x
#
#     def __iter__(self) :
#         for k in self.__nodes :
#             n = self.__nodes[k]
#             child_names = []
#             children = []
#             for x in n :
#                 if type(x) == str :
#                     x = self._deref(x)
#                     child_names.append(x)
#                 else :
#                     key = '%s_%s' % (k,len(child_names) )
#                     child_names.append( key )
#                     if type(x).__name__ != 'atom' :
#                         x_children = [self._deref(y) for y in x.children]
#                         children.append( (key, self._create_conj(x_children) ) )
#                     else :
#                         children.append( (key, x ))
#             if len(child_names) > 1 :
#                 yield (k, self._create_disj(child_names), 'disj')
#                 for i,c in children :
#                     yield (i, c, type(c).__name__)
#             else :
#                 for i,c in children :
#                     yield (k, c, type(c).__name__)
#
#     def __len__(self) :
#         return len(self.__nodes)
#
#     ##################################################################################
#     ####                            OUTPUT GENERATION                             ####
#     ##################################################################################
#
#     def toLogicFormula(self) :
#         target = LogicFormula(auto_compact=False)
#         translate = {}
#         i = 0
#         for k,n,t in self :
#             i += 1
#             translate[k] = i
#             translate['-' + str(k) ] = -i
#         for k,n,t in self :
#             if t == 'atom' :
#                 i = target.add_atom( n.identifier, n.probability, n.group )
#             elif t == 'disj' :
#                 i = target.add_or( [ translate[x] for x in n.children ] )
#             elif t == 'conj' :
#                 i = target.add_and( [ translate[x] for x in n.children ] )
#             assert(i == translate[k])
#
#         for name, key, label in self.get_names_with_label():
#             key = self._deref(key)
#             target.add_name(name, translate[key], label)
#
#         return target
#
#     @classmethod
#     def loadFrom(cls, lp) :
#         interm = StringKeyLogicFormula()
#         for c in lp :
#             if type(c).__name__ == 'Clause' :
#                 key = str(c.head)
#                 body = []
#                 current = c.body
#                 while type(current).__name__ == 'And' :
#                     if type(current.op1).__name__ == 'Not' :
#                         body.append('-' + str(current.op1.child))
#                     else :
#                         body.append(str(current.op1))
#                     current = current.op2
#                 if type(current).__name__ == 'Not' :
#                     body.append('-' + str(current.child))
#                 else :
#                     body.append(str(current))
#                 interm.add_and( body, key=key )
#                 interm.add_name(key, key, interm.LABEL_NAMED)
#             elif type(c).__name__ == 'Term' :
#                 key = str(c.with_probability())
#                 interm.add_atom( key, c.probability, None )
#                 interm.add_name(key, key, interm.LABEL_NAMED)
#             else :
#                 raise Exception("Unexpected type: '%s'" % type(c).__name__)
#         return interm
#


# Alternative cycle breaking below: loop formula's
#   ASSAT: Computing Answer Sets of A Logic Program By SAT Solvers
#   Fangzhen Lin and Yuting Zhao
#   AAAI'02
# Not in use:
#   - does not work with SDD
#   - added constraints can become extremely large

# #@transform(LogicFormula, LogicDAG)
# def breakCyclesConstraint(source, target) :
#     relevant = [False] * (len(source)+1)
#     cycles = {}
#     for name, node, label in source.getNamesWithLabel() :
#         if label != interm.LABEL_NAMED :
#             for c_in in findCycles( source, node, [], relevant) :
#                 c_in = tuple(sorted(c_in))
#                 if c_in in cycles :
#                     pass
#                 else :
#                     cycles[c_in] = splitCycle(source, c_in)
#
#     for c_in, c_out in cycles.items() :
#         source.addConstraint(ConstraintLoop(c_in, c_out))
#     return source
#
# def splitCycle(src, loop) :
#     cycle_free = []
#     for n in loop :
#         n = src.get_node(n)
#         t = type(n).__name__
#         if t == 'disj' :
#             cycle_free += [ c for c in n.children if not c in loop ]
#         elif t == 'conj' :
#             pass
#         else :
#             raise Exception('?')
#     return cycle_free
#
# def findCycles( src, a, path, relevant=None ) :
#     n = src.get_node(a)
#     t = type(n).__name__
#     if relevant != None : relevant[a] = True
#     try :
#         s = path.index(a)
#         yield path[s:]
#     except ValueError :
#         if t == 'atom' :
#             pass
#         else :
#             for c in n.children :
#                 for p in findCycles( src, c, path + [a], relevant ) :
#                     yield p
#
#
# class ConstraintLoop(Constraint) :
#     """Loop breaking constraint."""
#
#     def __init__(self, cycle_nodes, noncycle_nodes) :
#         self.in_loop = cycle_nodes
#         self.ex_loop = noncycle_nodes
#         self.in_node = None
#
#     def __str__(self) :
#         return 'loop_break(%s, %s)' % (list(self.in_loop), list(self.ex_loop))
#
#     def isTrue(self) :
#         return False
#
#     def isFalse(self) :
#         return False
#
#     def is_nontrivial(self) :
#         return True
#
#     def as_clauses(self) :
#         if self.is_nontrivial() :
#             ex_loop = tuple(self.ex_loop)
#             lines = []
#             for m in self.in_loop :
#                 lines.append( ex_loop + (-m,) )
#             return lines
#         else :
#             return []
#
#     def updateWeights(self, weights, semiring) :
#         """Update the weights of the logic formula accordingly."""
#         pass
#
#     def copy( self, rename={} ) :
#         cycle_nodes = set(rename.get(x,x) for x in self.in_loop)
#         noncycle_nodes = set(rename.get(x,x) for x in self.ex_loop)
#         result = ConstraintLoop( cycle_nodes, noncycle_nodes )
#         return result

#
# def copyFormula(source, target) :
#     for i, n, t in source :
#         if t == 'atom' :
#             target.add_atom( n.identifier, n.probability, n.group )
#         elif t == 'conj' :
#             target.add_and( n.children )
#         elif t == 'disj' :
#             target.add_or( n.children )
#         else :
#             raise TypeError("Unknown node type '%s'" % t)
#
#     for name, node, label in source.get_names_with_label():
#         target.add_name(name, node, label)
#
# def breakCycles_lp(source, target=None) :
#
#     if target != None :
#         copyFormula(source,target)
#     else :
#         target = source
#
#     tmp_file = tempfile.mkstemp('.lp')[1]
#     with open(tmp_file, 'w') as f :
#         lf_to_smodels(target, f)
#     output = subprocess.check_output(['lp2acyc', tmp_file])
#     smodels_to_lf( target, output )
#
#     try :
#         os.remove(tmp_file)
#     except OSError :
#         pass
#
#     return target
#
#
#
# def expand_node( formula, node_id ) :
#     """Expand conjunctions by their body until a disjunction or atom is encountered.
#     This method assumes that all cycles go through a disjunctive node.
#     """
#     node = formula.get_node(abs(node_id))
#     nodetype = type(node).__name__
#     conjuncts = []
#     if nodetype == 'disj' :
#         return [node_id]
#     elif nodetype == 'atom' :
#         return [node_id]
#     elif node_id < 0 :
#         return [node_id]
#     else : # conj
#         for c in node.children :
#             conjuncts += expand_node(formula,c)
#         return conjuncts
#
# def lf_to_smodels( formula, out ) :
#
#     # '1' is an internal atom => false
#
#     #print (formula)
#     # Write rules
#     # Basic rule:
#     #   1 head #lits #neglits [ body literals with negative first ]
#     for i,n,t in formula :
#         if t == 'disj' :
#             for c in n.children :
#                 body = expand_node(formula, c)
#                 l = len(body)
#                 nl = len([ b for b in body if b < 0 ])
#                 body = [ abs(b)+1 for b in sorted(body) ]
#                 print('1 %s %s %s %s' % (i+1, l, nl, ' '.join(map(str,sorted(body))) ), file=out)
#     print (0, file=out)
#
#     for i,n,t in formula :
#         if t == 'atom' or t == 'disj' :
#             print (i+1, i, file=out)
#
#     # Symbol table => must contain all (otherwise hidden in output)
#     # Facts and disjunctions
#     #   2 a
#     #   3 b
#
#     print (0, file=out)
#     print ('B+', file=out)
#     # B+  positive evidence?
#
#     print (0, file=out)
#     print ('B-', file=out)
#     # B-  negative evidence?
#
#     print (0, file=out)
#
#     # Number of models
#     print (1, file=out)
#
# def smodels_to_lf( formula, acyclic ) :
#
#     section = 0
#     rules = defaultdict(list)
#     data = [[]]
#     for line in acyclic.split('\n') :
#         if line == '0' :
#             section += 1
#             data.append([])
#         else :
#             data[-1].append( line )
#
#     acyc_nodes = frozenset([ int(x.split()[0]) for x in data[1] if '_acyc_' in x ])
#     given_nodes = frozenset([ int(x.split()[0]) for x in data[1] if not '_acyc_' in x ])
#     if len(data[3]) > 1 :
#         root_node = int(data[3][1])
#     else :
#         root_node = None
#
#     for line in data[0] :
#         line = line.split()
#         line_head = int(line[1])
#         line_neg = int(line[3])
#         line_body_neg = frozenset(map(int,line[4:4+line_neg]))
#         line_body_pos = frozenset(map(int,line[4+line_neg:]))
#
#         # acyc_nodes are true
#         if acyc_nodes & line_body_neg : continue
#         # part of original program
#         if line_head in given_nodes : continue
#         # acyc_nodes are true
#         line_body_pos -= acyc_nodes
#         body = sorted([ -a for a in line_body_neg ] + list(line_body_pos))
#         rules[line_head].append(body)
#
#     translate = {}
#     for head in rules :
#         acyc_insert(formula, rules, head, given_nodes, translate)
#
#     if root_node != None :
#         formula.addConstraint(TrueConstraint(translate[root_node]))
#
#     #print (formula)
#     # print (translate[root_node])
#
# def acyc_insert( formula, rules, head, given, translate ) :
#     if head < 0 :
#         f = -1
#     else :
#         f = 1
#
#     if abs(head) in translate :
#         return f * translate[abs(head)]
#     elif abs(head) in given :
#         return f * (abs(head)-1)
#     else :
#         disjuncts = []
#         for body in rules[abs(head)] :
#             new_body = [ acyc_insert(formula, rules, x, given, translate ) for x in body ]
#             disjuncts.append(formula.add_and( new_body ))
#         new_node = formula.add_or( disjuncts )
#         translate[abs(head)] = new_node
#         return f*new_node
#
# import collections
#
# class OrderedSet(collections.MutableSet):
#
#     def __init__(self, iterable=None):
#         self.end = end = []
#         end += [None, end, end]         # sentinel node for doubly linked list
#         self.map = {}                   # key --> [key, prev, next]
#         if iterable is not None:
#             self |= iterable
#
#     def __len__(self):
#         return len(self.map)
#
#     def __contains__(self, key):
#         return key in self.map
#
#     def add(self, key):
#         if key not in self.map:
#             end = self.end
#             curr = end[1]
#             curr[2] = end[1] = self.map[key] = [key, curr, end]
#
#     def discard(self, key):
#         if key in self.map:
#             key, prev, next = self.map.pop(key)
#             prev[2] = next
#             next[1] = prev
#
#     def __iter__(self):
#         end = self.end
#         curr = end[2]
#         while curr is not end:
#             yield curr[0]
#             curr = curr[2]
#
#     def __reversed__(self):
#         end = self.end
#         curr = end[1]
#         while curr is not end:
#             yield curr[0]
#             curr = curr[1]
#
#     def pop(self, last=True):
#         if not self:
#             raise KeyError('set is empty')
#         key = self.end[1][0] if last else self.end[2][0]
#         self.discard(key)
#         return key
#
#     def __repr__(self):
#         if not self:
#             return '%s()' % (self.__class__.__name__,)
#         return '%s(%r)' % (self.__class__.__name__, list(self))
#
#     def __eq__(self, other):
#         if isinstance(other, OrderedSet):
#             return len(self) == len(other) and list(self) == list(other)
#         return set(self) == set(other)



"""
problog.engine - Grounding engine
---------------------------------

Grounding engine to transform a ProbLog program into a propositional formula.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

import logging
import os

from collections import defaultdict, namedtuple

from .program import LogicProgram
from .logic import *
from .formula import LogicFormula
from .engine_unify import *

from .core import transform
from .errors import GroundingError
from .util import Timer


@transform(LogicProgram, LogicFormula)
def ground(model, target=None, queries=None, evidence=None, propagate_evidence=False, **kwdargs):
    """Ground a given model.

    :param model: logic program to ground
    :type model: LogicProgram
    :param target: formula in which to store ground program
    :type target: LogicFormula
    :param queries: list of queries to override the default
    :param evidence: list of evidence atoms to override the default
    :return: the ground program
    :rtype: LogicFormula
    """
    return DefaultEngine(**kwdargs).ground_all(model, target, queries=queries, evidence=evidence,
                                               propagate_evidence=propagate_evidence)


class GenericEngine(object):  # pragma: no cover
    """Generic interface to a grounding engine."""

    def prepare(self, db):
        """Prepare the given database for querying.
        Calling this method is optional.

       :param db: logic program
       :returns: logic program in optimized format where builtins are initialized and directives \
       have been evaluated
        """
        raise NotImplementedError('GenericEngine.prepare is an abstract method.')

    def query(self, db, term):
        """Evaluate a query without generating a ground program.

       :param db: logic program
       :param term: term to query; variables should be represented as None
       :returns: list of tuples of argument for which the query succeeds.
        """
        raise NotImplementedError('GenericEngine.query is an abstract method.')

    def ground(self, db, term, target=None, label=None):
        """Ground a given query term and store the result in the given ground program.

       :param db: logic program
       :param term: term to ground; variables should be represented as None
       :param target: target logic formula to store grounding in (a new one is created if none is \
       given)
       :param label: optional label (query, evidence, ...)
       :returns: logic formula (target if given)
        """
        raise NotImplementedError('GenericEngine.ground is an abstract method.')

    def ground_all(self, db, target=None, queries=None, evidence=None):
        """Ground all queries and evidence found in the the given database.

       :param db: logic program
       :param target: logic formula to ground into
       :param queries: list of queries to evaluate instead of the ones in the logic program
       :param evidence: list of evidence to evaluate instead of the ones in the logic program
       :returns: ground program
        """
        raise NotImplementedError('GenericEngine.ground_all is an abstract method.')


class ClauseDBEngine(GenericEngine):
    """Parent class for all Python ClauseDB-based engines."""

    UNKNOWN_ERROR = 0
    UNKNOWN_FAIL = 1

    def __init__(self, builtins=True, **kwdargs):
        self.__builtin_index = {}
        self.__builtins = []
        self.__externals = {}

        self._unique_number = 0
        self.unknown = self.UNKNOWN_ERROR

        if builtins:
            self.load_builtins()

        self.functions = {}

    def load_builtins(self):
        """Load default builtins."""
        raise NotImplementedError("ClauseDBEngine.loadBuiltIns is an abstract function.")

    def get_builtin(self, index):
        """Get builtin's evaluation function based on its identifier.
       :param index: index of the builtin
       :return: function that evaluates the builtin
        """
        real_index = -(index + 1)
        return self.__builtins[real_index]

    def add_builtin(self, predicate, arity, function):
        """Add a builtin.

        :param predicate: name of builtin predicate
        :param arity: arity of builtin predicate
        :param function: function to execute builtin
        """
        sig = '%s/%s' % (predicate, arity)
        self.__builtin_index[sig] = -(len(self.__builtins) + 1)
        self.__builtins.append(function)

    def get_builtins(self):
        """Get the list of builtins."""
        return self.__builtin_index

    def prepare(self, db):
        """Convert given logic program to suitable format for this engine.
        Calling this method is optional, but it allows to perform multiple operations on the same \
        database.
        This also executes any directives in the input model.

        :param db: logic program to prepare for evaluation
        :type db: LogicProgram
        :return: logic program in a suitable format for this engine
        :rtype: ClauseDB
        """
        result = ClauseDB.createFrom(db, builtins=self.get_builtins())
        self._process_directives(result)
        return result

    def call(self, query, database, target, transform=None, **kwdargs):
        raise NotImplementedError("ClauseDBEngine.call is an abstract function.")

    def execute(self, node_id, database=None, context=None, target=None, **kwdargs):
        raise NotImplementedError("ClauseDBEngine.execute is an abstract function.")

    def get_non_cache_functor(self):
        """Get a unique functor that is excluded from caching.

        :return: unique functor that is excluded from caching
        :rtype: basestring
        """
        self._unique_number += 1
        return '_nocache_%s' % self._unique_number

    def _process_directives(self, db):
        """Process directives present in the database."""
        term = Term('_directive')
        directive_node = db.find(term)
        if directive_node is None:
            return True    # no directives
        directives = db.get_node(directive_node).children

        gp = LogicFormula()
        while directives:
            current = directives.pop(0)
            self.execute(current, database=db, context=self.create_context((), define=None),
                         target=gp)
        return True

    # noinspection PyUnusedLocal
    def create_context(self, content, define=None):
        """Create a variable context."""
        return content

    def _fix_context(self, context):
        return tuple(context)

    def _clone_context(self, context):
        return list(context)

    def query(self, db, term, **kwdargs):
        """

       :param db:
       :param term:
       :param kwdargs:
       :return:
        """
        gp = LogicFormula()
        if term.is_negated():
            term = -term
            negative = True
        else:
            negative = False
        gp, result = self._ground(db, term, gp, **kwdargs)
        if negative:
            if not result:
                return [term]
            else:
                return []
        else:
            return [x for x, y in result]

    def ground(self, db, term, target=None, label=None, **kwdargs):
        """Ground a query on the given database.

       :param db: logic program
       :type db: LogicProgram
       :param term: query term
       :type term: Term
       :param gp: output data structure (for incremental grounding)
       :type gp: LogicFormula
       :param label: type of query (e.g. ``query``, ``evidence`` or ``-evidence``)
       :type label: str
       :param kwdargs: additional arguments
       :return: ground program containing the query
       :rtype: LogicFormula
        """
        if term.is_negated():
            negated = True
            term = -term
        else:
            negated = False

        target, results = self._ground(db, term, target, silent_fail=False, **kwdargs)
        for args, node_id in results:
            term_store = term.with_args(*args)
            if negated:
                target.add_name(-term_store, -node_id, label)
            else:
                target.add_name(term_store, node_id, label)
        if not results:
            target.add_name(term, None, label)
        return target

    def _ground(self, db, term, gp=None, silent_fail=True, assume_prepared=False, **kwdargs):
        """

        :param db:
        :type db: LogicProgram
        :param term:
        :param gp:
        :param silent_fail:
        :param assume_prepared:
        :param kwdargs:
        :return:
        """
        # Convert logic program if needed.
        if not assume_prepared:
            db = self.prepare(db)
        # Create a new target datastructure if none was given.
        if gp is None:
            gp = LogicFormula()
        # Find the define node for the given query term.
        clause_node = db.find(term)
        # If term not defined: fail query (no error)    # TODO add error to make it consistent?
        if clause_node is None:
            # Could be builtin?
            clause_node = db.get_builtin(term.signature)
        if clause_node is None:
            if silent_fail or self.unknown == self.UNKNOWN_FAIL:
                return gp, []
            else:
                raise UnknownClause(term.signature, location=db.lineno(term.location))

        try:
            context = self.create_context(term.args)
            context, xxx = substitute_call_args(context, context)
            results = self.execute(clause_node, database=db, target=gp, context=context, **kwdargs)
        except UnknownClauseInternal:
            if silent_fail or self.unknown == self.UNKNOWN_FAIL:
                return gp, []
            else:
                raise UnknownClause(term.signature, location=db.lineno(term.location))

        return gp, results

    def ground_evidence(self, db, target, evidence):
        logger = logging.getLogger('problog')
        # Ground evidence
        for query in evidence:
            if len(query) == 1:  # evidence/1
                if query[0].is_negated():
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(db, -query[0], target, label=target.LABEL_EVIDENCE_NEG, is_root=True)
                    logger.debug("Ground program size: %s", len(target))
                else:
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(db, query[0], target, label=target.LABEL_EVIDENCE_POS, is_root=True)
                    logger.debug("Ground program size: %s", len(target))
            else:  # evidence/2
                if str(query[1]) == 'true' or query[1] == True:
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(db, query[0], target, label=target.LABEL_EVIDENCE_POS, is_root=True)
                    logger.debug("Ground program size: %s", len(target))
                elif str(query[1]) == 'false' or query[1] == False:
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(db, query[0], target, label=target.LABEL_EVIDENCE_NEG, is_root=True)
                    logger.debug("Ground program size: %s", len(target))
                else:
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(db, query[0], target, label=target.LABEL_EVIDENCE_MAYBE, is_root=True)
                    logger.debug("Ground program size: %s", len(target))

    def ground_queries(self, db, target, queries):
        logger = logging.getLogger('problog')
        for query in queries:
            logger.debug("Grounding query '%s'", query)
            target = self.ground(db, query, target, label=target.LABEL_QUERY)
            logger.debug("Ground program size: %s", len(target))

    def ground_all(self, db, target=None, queries=None, evidence=None, propagate_evidence=False):
        db = self.prepare(db)
        logger = logging.getLogger('problog')
        with Timer('Grounding'):
            # Load queries: use argument if available, otherwise load from database.
            if queries is None:
                queries = [q[0] for q in self.query(db, Term('query', None))]
            for query in queries:
                if not isinstance(query, Term):
                    raise GroundingError('Invalid query')   # TODO can we add a location?
            # Load evidence: use argument if available, otherwise load from database.
            if evidence is None:
                evidence = self.query(db, Term('evidence', None, None))
                evidence += self.query(db, Term('evidence', None))

            for ev in evidence:
                if not isinstance(ev[0], Term):
                    raise GroundingError('Invalid evidence')   # TODO can we add a location?

            # Initialize target if not given.
            if target is None:
                target = LogicFormula()
            # Ground queries
            if propagate_evidence:
                with Timer('Propagating evidence'):
                    self.ground_evidence(db, target, evidence)
                    # delattr(target, '_cache')
                    target.lookup_evidence = {}
                    ev_nodes = [node for name, node in target.evidence()
                                if node != 0 and node is not None]
                    target.propagate(ev_nodes, target.lookup_evidence)
                    self.ground_queries(db, target, queries)
                logger.debug('Propagated evidence: %s' % list(target.lookup_evidence))
            else:
                self.ground_queries(db, target, queries)
                self.ground_evidence(db, target, evidence)
        return target

    def add_external_calls(self, externals):
        self.__externals.update(externals)

    def get_external_call(self, func_name):
        if self.__externals is None or func_name not in self.__externals:
            return None
        return self.__externals[func_name]


class UnknownClauseInternal(Exception):
    """Undefined clause in call used internally."""
    pass


class NonGroundProbabilisticClause(GroundingError):
    """Encountered a non-ground probabilistic clause."""

    def __init__(self, location):
        GroundingError.__init__(self, 'Encountered a non-ground probabilistic clause', location)


class UnknownClause(GroundingError):
    """Undefined clause in call."""

    def __init__(self, signature, location):
        GroundingError.__init__(self, "No clauses found for '%s'" % signature, location)
        self.signature = signature


from .engine_stack import StackBasedEngine as DefaultEngine


def intersection(l1, l2):
    i = 0
    j = 0
    n1 = len(l1)
    n2 = len(l2)
    r = []
    a = r.append
    while i < n1 and j < n2:
        if l1[i] == l2[j]:
            a(l1[i])
            i += 1
            j += 1
        elif l1[i] < l2[j]:
            i += 1
        else:
            j += 1
    return r


class ClauseIndex(list):

    def __init__(self, parent, arity):
        list.__init__(self)
        self.__parent = parent
        self.__basetype = OrderedSet
        self.__index = [defaultdict(self.__basetype) for _ in range(0, arity)]
        self.__optimized = False

    def find(self, arguments):
        results = None
        for i, arg in enumerate(arguments):
            if not is_ground(arg):
                pass  # Variable => no restrictions
            else:
                curr = self.__index[i].get(arg)
                none = self.__index[i].get(None, self.__basetype())
                if curr is None:
                    curr = none
                else:
                    curr |= none

                if results is None:  # First argument with restriction
                    results = curr
                else:
                    results = results & curr       # for some reason &= doesn't work here
            if results is not None and not results:
                return []
        if results is None:
            return self
        else:
            return results

    def _add(self, key, item):
        assert not self.__optimized
        for i, k in enumerate(key):
            self.__index[i][k].add(item)

    def append(self, item):
        list.append(self, item)
        key = []
        args = self.__parent.get_node(item).args
        for arg in args:
            if is_ground(arg):
                key.append(arg)
            else:
                key.append(None)
        self._add(key, item)


class ClauseDB(LogicProgram):
    """Compiled logic program.

    A logic program is compiled into a table of instructions.
    The types of instructions are:

    define( functor, arity, defs )
        Pointer to all definitions of functor/arity.
        Definitions can be: ``fact``, ``clause`` or ``adc``.

    clause( functor, arguments, bodynode, varcount )
        Single clause. Functor is the head functor, Arguments are the head arguments. Body node is a pointer to the node representing the body. Var count is the number of variables in head and body.

    fact( functor, arguments, probability )
        Single fact.

    adc( functor, arguments, bodynode, varcount, parent )
        Single annotated disjunction choice. Fields have same meaning as with ``clause``, parent_node points to the parent ``ad`` node.

    ad( childnodes )
        Annotated disjunction group. Child nodes point to the ``adc`` nodes of the clause.

    call( functor, arguments, defnode )
        Body literal with call to clause or builtin. Arguments contains the call arguments, definition node is the pointer to the definition node of the given functor/arity.

    conj( childnodes )
        Logical and. Currently, only 2 children are supported.

    disj( childnodes )
        Logical or. Currently, only 2 children are supported.

    neg( childnode )
        Logical not.

    """
    _define = namedtuple('define', ('functor', 'arity', 'children', 'location'))
    _clause = namedtuple('clause', ('functor', 'args', 'probability', 'child',
                                    'varcount', 'locvars', 'group', 'location'))
    _fact = namedtuple('fact', ('functor', 'args', 'probability', 'location'))
    _call = namedtuple('call', ('functor', 'args', 'defnode', 'location', 'op_priority', 'op_spec'))
    _disj = namedtuple('disj', ('children', 'location'))
    _conj = namedtuple('conj', ('children', 'location'))
    _neg = namedtuple('neg', ('child', 'location'))
    _choice = namedtuple('choice', ('functor', 'args', 'probability', 'locvars', 'group', 'choice', 'location'))
    _extern = namedtuple('extern', ('functor', 'arity', 'function',))

    FUNCTOR_CHOICE = 'choice'
    FUNCTOR_BODY = 'body'

    def __init__(self, builtins=None, parent=None):
        LogicProgram.__init__(self)
        self.__nodes = []   # list of nodes
        self.__heads = {}   # head.sig => node index

        self.__builtins = builtins

        self.data = {}

        self.__parent = parent
        if parent is None:
            self.__offset = 0
        else:
            if hasattr(parent, 'line_info'):
                self.line_info = parent.line_info
            if hasattr(parent, 'source_files'):
                self.source_files = parent.source_files[:]
            self.__offset = len(parent)

    def __len__(self):
        return len(self.__nodes) + self.__offset

    def extend(self):
        return ClauseDB(parent=self)

    def set_data(self, key, value):
        self.data[key] = value

    def update_data(self, key, value):
        if self.has_data(key):
            if type(value) == list:
                self.data[key] += value
            elif type(value) == dict:
                self.data[key].update(value)
            else:
                raise TypeError('Can\'t update data of type \'%s\'' % type(value))
        else:
            self.data[key] = value

    def has_data(self, key):
        return key in self.data

    def get_data(self, key, default=None):
        return self.data.get(key, default)

    def get_builtin(self, signature):
        if self.__builtins is None:
            if self.__parent is not None:
                return self.__parent.get_builtin(signature)
            else:
                return None
        else:
            return self.__builtins.get(signature)

    def get_reserved_names(self):
        return {self.FUNCTOR_CHOICE, self.FUNCTOR_BODY}

    def is_reserved_name(self, name):
        return name is self.get_reserved_names()

    def _create_index(self, arity):
        # return []
        return ClauseIndex(self, arity)

    def _add_and_node(self, op1, op2, location=None):
        """Add an *and* node."""
        return self._append_node(self._conj((op1, op2), location))

    def _add_not_node(self, op1, location=None):
        """Add a *not* node."""
        return self._append_node(self._neg(op1, location))

    def _add_or_node(self, op1, op2, location=None):
        """Add an *or* node."""
        return self._append_node(self._disj((op1, op2), location))

    def _add_define_node(self, head, childnode):
        define_index = self._add_head(head)
        define_node = self.get_node(define_index)
        if not define_node:
            clauses = self._create_index(head.arity)
            self._set_node(define_index, self._define(head.functor, head.arity, clauses, head.location))
        else:
            clauses = define_node.children
        clauses.append(childnode)
        return childnode

    def _add_choice_node(self, choice, functor, args, probability, locvars, group, location=None):
        choice_node = self._append_node(self._choice(functor, args, probability, locvars, group, choice, location))
        return choice_node

    def _add_clause_node(self, head, body, varcount, locvars, group=None):
        clause_node = self._append_node(self._clause(
            head.functor, head.args, head.probability, body, varcount, locvars, group, head.location))
        return self._add_define_node(head, clause_node)

    def _add_call_node(self, term):
        """Add a *call* node."""
        if term.signature in ('query/1', 'evidence/1', 'evidence/2'):
            raise AccessError("Can\'t call %s directly." % term.signature)

        defnode = self._add_head(term, create=False)
        return self._append_node(self._call(term.functor, term.args, defnode, term.location,
                                            term.op_priority, term.op_spec))

    def get_node(self, index):
        """Get the instruction node at the given index.

        :param index: index of the node to retrieve
        :type index: :class:`int`
        :returns: requested node
        :rtype: :class:`tuple`
        :raises IndexError: the given index does not point to a node
        """
        if index < self.__offset:
            return self.__parent.get_node(index)
        else:
            return self.__nodes[index - self.__offset]

    def _set_node(self, index, node):
        if index < self.__offset:
            raise IndexError('Can\'t update node in parent.')
        else:
            self.__nodes[index - self.__offset] = node

    def _append_node(self, node=()):
        index = len(self)
        self.__nodes.append(node)
        return index

    def _get_head(self, head):
        node = self.__heads.get(head.signature)
        if node is None and self.__parent:
            node = self.__parent._get_head(head)
        return node

    def _set_head(self, head, index):
        self.__heads[head.signature] = index

    def _add_head(self, head, create=True):
        if self.is_reserved_name(head.functor):
            raise AccessError("'%s' is a reserved name" % head.functor)
        node = self.get_builtin(head.signature)
        if node is not None:
            if create:
                raise AccessError("Can not overwrite built-in '%s'." % head.signature)
            else:
                return node

        node = self._get_head(head)
        if node is None:
            if create:
                node = self._append_node(self._define(head.functor, head.arity, self._create_index(head.arity),
                                                      head.location))
            else:
                node = self._append_node()
            self._set_head(head, node)
        return node

    def find(self, head):
        """Find the ``define`` node corresponding to the given head.

        :param head: clause head to match
        :type head: :class:`.basic.Term`
        :returns: location of the clause node in the database, \
                     returns ``None`` if no such node exists
        :rtype: :class:`int` or ``None``
        """
        return self._get_head(head)

    def __repr__(self):
        s = ''
        for i, n in enumerate(self.__nodes):
            i += self.__offset
            s += '%s: %s\n' % (i, n)
        s += str(self.__heads)
        return s

    def add_clause(self, clause):
        """Add a clause to the database.

       :param clause: Clause to add
       :type clause: Clause
       :returns: location of the definition node in the database
       :rtype: int
        """
        return self._compile(clause)

    def add_fact(self, term):
        """Add a fact to the database.
       :param term: fact to add
       :type term: Term
       :return: position of the definition node in the database
       :rtype: int
        """

        # Count the number of variables in the fact
        variables = _AutoDict()
        term.apply(variables)
        # If the fact has variables, threat is as a clause.
        if len(variables) == 0:
            fact_node = self._append_node(self._fact(term.functor, term.args,
                                                     term.probability, term.location))
            return self._add_define_node(term, fact_node)
        else:
            return self.add_clause(Clause(term, Term('true')))

    def add_extern(self, predicate, arity, function):
        head = Term(predicate, *[None] * arity)
        node_id = self._get_head(head)
        if node_id is None:
            node_id = self._append_node(self._extern(predicate, arity, function))
            self._set_head(head, node_id)
        else:
            node = self.get_node(node_id)
            if node == ():
                self._set_node(node_id, self._extern(predicate, arity, function))
            else:
                raise AccessError("External function overrides already defined predicate '%s'"
                                  % head.signature)

    def get_local_scope(self, signature):
        if signature == 'findall/3':
            return 0, 1
        else:
            return []

    def _compile(self, struct, variables=None):
        """Compile the given structure and add it to the database.

        :param struct: structure to compile
        :type struct: Term
        :param variables: mapping between variable names and variable index
        :type variables: _AutoDict
        :return: position of the compiled structure in the database
        :rtype: int
        """
        if variables is None:
            variables = _AutoDict()

        if isinstance(struct, And):
            op1 = self._compile(struct.op1, variables)
            op2 = self._compile(struct.op2, variables)
            return self._add_and_node(op1, op2)
        elif isinstance(struct, Or):
            op1 = self._compile(struct.op1, variables)
            op2 = self._compile(struct.op2, variables)
            return self._add_or_node(op1, op2)
        elif isinstance(struct, Not):
            child = self._compile(struct.child, variables)
            return self._add_not_node(child, location=struct.location)
        elif isinstance(struct, Term) and struct.signature == 'not/1':
            child = self._compile(struct.args[0], variables)
            return self._add_not_node(child, location=struct.location)
        elif isinstance(struct, AnnotatedDisjunction):
            # Determine number of variables in the head
            new_heads = [head.apply(variables) for head in struct.heads]

            # Group id
            group = len(self.__nodes)

            # Create the body clause
            body_node = self._compile(struct.body, variables)
            body_count = len(variables)
            # Body arguments
            body_args = tuple(range(0, len(variables)))
            body_functor = self.FUNCTOR_BODY + '_' + str(len(self))
            if len(new_heads) > 1:
                heads_list = Term('multi')  # list2term(new_heads)
            else:
                heads_list = new_heads[0]
            body_head = Term(body_functor, Constant(group), heads_list, *body_args)
            self._add_clause_node(body_head, body_node, len(variables), variables.local_variables)
            clause_body = self._add_head(body_head)
            for choice, head in enumerate(new_heads):
                # For each head: add choice node
                choice_functor = Term(self.FUNCTOR_CHOICE,
                                      Constant(group), Constant(choice), head.with_probability())
                choice_node = self._add_choice_node(choice, choice_functor, body_args,
                                                    head.probability, variables.local_variables,
                                                    group, head.location)
                choice_call = self._append_node(self._call(choice_functor, body_args, choice_node,
                                                           head.location, None, None))
                body_call = self._append_node(self._call(body_functor, body_head.args, clause_body,
                                                         head.location, None, None))
                choice_body = self._add_and_node(body_call, choice_call)
                self._add_clause_node(head, choice_body, body_count, {}, group=group)
            return None
        elif isinstance(struct, Clause):
            if struct.head.probability is not None:
                return self._compile(AnnotatedDisjunction([struct.head], struct.body))
            else:
                new_head = struct.head.apply(variables)
                body_node = self._compile(struct.body, variables)
                return self._add_clause_node(new_head, body_node, len(variables),
                                             variables.local_variables)
        elif isinstance(struct, Var):
            return self._add_call_node(Term('call', struct.apply(variables),
                                            location=struct.location))
        elif isinstance(struct, Term):
            local_scope = self.get_local_scope(struct.signature)
            if local_scope:
                # Special case for findall: any variables added by the first
                #  two arguments of findall are 'local' variables.
                args = []
                for i, a in enumerate(struct.args):
                    if not isinstance(a, Term):
                        # For nested findalls: 'a' can be a raw variable pointer
                        # Temporarily wrap it in a Term, so we can call 'apply' on it.
                        a = Term('_', a)
                    if i in local_scope:
                        variables.enter_local()
                        new_arg = a.apply(variables)
                        variables.exit_local()
                    else:
                        new_arg = a.apply(variables)
                    if a.functor == '_':
                        # If the argument was temporarily wrapped: unwrap it.
                        new_arg = new_arg.args[0]
                    args.append(new_arg)
                return self._add_call_node(struct(*args))
            else:
                return self._add_call_node(struct.apply(variables))
        else:
            raise ValueError("Unknown structure type: '%s'" % struct)

    def _create_vars(self, term):
        if type(term) == int:
            return Var('V_' + str(term))
        else:
            args = [self._create_vars(arg) for arg in term.args]
            return term.with_args(*args)

    def _extract(self, node_id):
        node = self.get_node(node_id)
        if not node:
            raise ValueError("Unexpected empty node.")

        nodetype = type(node).__name__
        if nodetype == 'fact':
            return Term(node.functor, *node.args, p=node.probability)
        elif nodetype == 'call':
            func = node.functor
            args = node.args
            if isinstance(func, Term):
                return self._create_vars(func(*(func.args + args)))
            else:
                return self._create_vars(Term(func, *args,
                                              priority=node.op_priority, opspec=node.op_spec))
        elif nodetype == 'conj':
            a, b = node.children
            return And(self._extract(a), self._extract(b))
        elif nodetype == 'disj':
            a, b = node.children
            return Or(self._extract(a), self._extract(b))
        elif nodetype == 'neg':
            return Not('\+', self._extract(node.child))
        else:
            raise ValueError("Unknown node type: '%s'" % nodetype)

    def __iter__(self):
        clause_groups = defaultdict(list)
        for index, node in enumerate(self.__nodes):
            index += self.__offset
            if not node:
                continue
            nodetype = type(node).__name__
            if nodetype == 'fact':
                yield Term(node.functor, *node.args, p=node.probability)
            elif nodetype == 'clause':
                if node.group is None:
                    head = self._create_vars(Term(node.functor, *node.args, p=node.probability))
                    yield Clause(head, self._extract(node.child))
                else:
                    clause_groups[node.group].append(index)
        for group in clause_groups.values():
            heads = []
            body = None
            for index in group:
                node = self.get_node(index)
                heads.append(self._create_vars(Term(node.functor, *node.args, p=node.probability)))
                if body is None:
                    body_node = self.get_node(node.child)
                    body_node = self.get_node(body_node.children[0])
                    body = self._create_vars(Term(body_node.functor, *body_node.args))
            yield AnnotatedDisjunction(heads, body)

    def iter_raw(self):
        """Iterate over clauses of model as represented in the database i.e. with choice facts and
         without annotated disjunctions.
        """

        clause_groups = defaultdict(list)
        for index, node in enumerate(self.__nodes):
            index += self.__offset
            if not node:
                continue
            nodetype = type(node).__name__
            if nodetype == 'fact':
                yield Term(node.functor, *node.args, p=node.probability)
            elif nodetype == 'clause':
                if node.group is None:
                    head = self._create_vars(Term(node.functor, *node.args, p=node.probability))
                    yield Clause(head, self._extract(node.child))
                else:
                    head = self._create_vars(Term(node.functor, *node.args))
                    yield Clause(head, self._extract(node.child))
            elif nodetype == 'choice':
                group = node.functor.args[0]
                c = node.functor(*(node.functor.args + node.args))
                clause_groups[group].append(c)
                yield c.with_probability(node.probability)

        for group in clause_groups.values():
            if len(group) > 1:
                yield Term('mutual_exclusive', list2term(group))

    def resolve_filename(self, filename):
        root = self.source_root
        if hasattr(filename, 'location') and filename.location:
            source_root = self.source_files[filename.location[0]]
            if source_root:
                root = os.path.dirname(source_root)

        atomstr = str(filename)
        if atomstr[0] == atomstr[-1] == "'":
            atomstr = atomstr[1:-1]
        filename = os.path.join(root, atomstr)
        return filename


class AccessError(GroundingError):
    pass


class _AutoDict(dict):

    def __init__(self):
        dict.__init__(self)
        self.__record = set()
        self.__anon = 0
        self.__localmode = False
        self.local_variables = set()

    def enter_local(self):
        self.__localmode = True

    def exit_local(self):
        self.__localmode = False

    def __getitem__(self, key):
        if key == '_' and self.__localmode:
            key = '_#%s' % self.__anon

        if key == '_':
            value = len(self)
            self.__anon += 1
            return value
        else:
            value = self.get(key)
            if value is None:
                value = len(self)
                self[key] = value
                if self.__localmode:
                    self.local_variables.add(value)
            elif not self.__localmode and value in self.local_variables:
                # Variable initially defined in local scope is reused outside local scope.
                # This means it's not local anymore.
                self.local_variables.remove(value)
            self.__record.add(value)
            return value

    def __len__(self):
        return dict.__len__(self) + self.__anon

    def usedVars(self):
        result = set(self.__record)
        self.__record.clear()
        return result

    def define(self, key):
        if key not in self:
            value = len(self)
            self[key] = value
